"""
The :mod:`py_fatigue.msc.corrections` module collects all the
mean stress corrections implemented in py_fatigue.
classes.
"""

from __future__ import annotations
import logging
import warnings

import matplotlib.pyplot as plt
import numpy as np

# from scipy.optimize import fsolve

from ..utils import make_axes, compile_specialized_newton, numba_newton
from ..styling import py_fatigue_formatwarning

warnings.formatwarning = py_fatigue_formatwarning


def dnvgl_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    detail_factor: float | int = 0.8,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to par.
    2.5 of`DNVGL-RP-C203 <https://bit.ly/3dUZ1OY>`_ which includes
    an attenuation factor :math:`p` for the stress ranges if the
    following cases:

    * base material without significant residual stresses \
    :math:`\\rightarrow p = 0.6`. This option neglects fully \
    compressive cycles.
    * welded connections without significant residual stresses \
    :math:`\\rightarrow p = 0.8`. This option multiplies the \
    stress range of fully compressive cycles by p.

    Given that the stress ranges are :math:`\\Delta \\sigma`,
    the corrected stress ranges are:

    .. math::

        \\Delta \\sigma_{corr} = f_m \\cdot \\Delta \\sigma

    where:

    .. math::

        f_m = \\frac{\\sigma_{max} + p \\cdot \\vert \\sigma_{min}
        \\vert}{\\sigma_{max} + \\vert\\sigma_{min} \\vert}

    with:

    .. math::

        \\frac{\\sigma_{amp}}{\\vert \\sigma_m \\vert} \\leq 1, \\quad
        \\sigma_{amp} = \\frac{\\sigma_{max} - \\sigma_{min}}{2} \\,\\land
        \\, \\sigma_m = \\frac{\\sigma_{max} + \\sigma_{min}}{2}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_m`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_{amp}`).
    detail_factor : float, optional
        mean stress attenuation factor (:math:`p`), defaults to 0.8
        in welded connections without significant residual stresses.
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges
        (:math:`\\Delta \\sigma_{corr}`).
    """

    max_stress = mean_stress + stress_amplitude
    min_stress = mean_stress - stress_amplitude

    f_m = (max_stress + detail_factor * np.abs(min_stress)) / (
        max_stress + np.abs(min_stress)
    )

    f_m[f_m > 1] = 1
    f_m[min_stress > 0] = 1
    if detail_factor == 0.8:
        f_m[f_m < detail_factor] = detail_factor
    elif detail_factor == 0.6:
        f_m[f_m < detail_factor] = 0
    else:
        raise ValueError(
            f"Detail factor = {str(detail_factor)} is not allowed in "
            + "DNVGL-RP-C203. Only 0.6 and 0.8 are permitted."
        )

    stress_range_corr = 2 * f_m * stress_amplitude

    # NOTE: next plot has been included for testing purposes
    if plot:  # pragma: no cover
        # from matplotlib.ticker import MaxNLocator
        _, axes = make_axes()
        xp, yp = zip(*sorted(zip(min_stress / max_stress, f_m)))
        axes.plot(
            xp,
            yp,
            linestyle=":",
            linewidth=1,
            color="#000000",
            marker=".",
            markersize=6,
            markerfacecolor="#000000",
            markeredgewidth=0.5,
            markeredgecolor="#666666",
        )
        plt.rcParams.update({"mathtext.default": "regular"})

        axes.set_xlabel("load ratio, $\\sigma_{min}/ \\sigma_{max}$")
        axes.set_ylabel("$f_m$")
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)
        axes.set_xlim((-20, 20))

    return stress_range_corr


def walker_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    gamma: float = 0.5,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to Walker model.

    The correction is given by:

    .. math::

        \\Delta \\sigma_{corr} = {\\sigma_{max}} ^ {(1 - \\gamma)} \\,
        \\sigma_{amp} ^ {\\gamma}

    with:

    .. math::

        \\sigma_{max} = \\sigma_{mean} + \\sigma_{amp}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_{mean}`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_{amp}`).
    gamma : float, optional
        The gamma Walker exponent (:math:`\\gamma`), defaults to 0.5.
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges (:math:`\\Delta \\sigma_{corr}`).
    """

    max_stress = mean_stress + stress_amplitude

    stress_range_corr = np.nan_to_num(
        (max_stress ** (1 - gamma)) * (stress_amplitude**gamma)
    )

    # NOTE: next plot has been included for testing purposes
    if plot:  # pragma: no cover
        # from matplotlib.ticker import MaxNLocator
        _, axes = make_axes()
        xp, yp = zip(*sorted(zip(2 * stress_amplitude, stress_range_corr)))
        axes.plot(
            xp,
            yp,
            linestyle=":",
            linewidth=1,
            color="#000000",
            marker=".",
            markersize=6,
            markerfacecolor="#000000",
            markeredgewidth=0.5,
            markeredgecolor="#666666",
        )
        plt.rcParams.update({"mathtext.default": "regular"})

        axes.set_xlabel("uncorrected stress range, $\\Delta \\sigma$")
        axes.set_ylabel("MSC stress range, $\\Delta \\sigma_{MSC}$")
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)

    return stress_range_corr


def swt_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to
    Smith-Watson-Topper model. It is equivalent to the Walker model
    with gamma = 0.5.

    The correction is given by:

    .. math::

        \\Delta \\sigma_{corr} = \\sqrt{\\sigma_{max} \\cdot \\sigma_{amp}}

    with:

    .. math::

        \\sigma_{max} = \\sigma_{mean} + \\sigma_{amp}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_{mean}`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_{amp}`).
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges (:math:`\\Delta \\sigma_{corr}`).
    """

    return walker_mean_stress_correction(
        mean_stress, stress_amplitude, gamma=0.5, plot=plot
    )


def goodman_haigh_mean_stress_correction(  # pylint: disable=R0912 # noqa: C901,E501
    amp_in: np.ndarray | list[float],
    mean_in: np.ndarray | list[float],
    r_out: float | np.ndarray | list[float],
    ult_s: float,
    correction_exponent: float,
    initial_guess: np.ndarray | None = None,
    plot: bool = False,
    logger: logging.Logger | None = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    This function extends the standard Goodman-Haigh correction formula which
    computes the corrected stress amplitude at zero mean stress (:math:`R=-1`)
    for a given set of input stress amplitudes and mean stresses.

    .. math::

        \\sigma_{amp, \\text{out}} =  \\frac{\\sigma_{amp, \\text{in}}}
        {1 - \\left(\\frac{\\sigma_{mean, \\text{in}}}{\\sigma_{ult}}
        \\right)^{n}}

    The correction exponent :math:`n` is a material property that can
    be obtained from experimental data.

    The current implementation extends the standard Goodman-Haigh formula by
    correcting the stress amplitude from an initial load ratio :math:`R_{in}`
    to any other load ratio :math:`R_{out}`, as long as the material's
    ultimate strength is known.
    The corrected stress amplitude :math:`\\sigma_{amp, \\text{out}}` is
    computed using the following implicit nonlinear equation:

    .. math::

        \\sigma_{amp, \\text{out}} =  \\frac{1}{ \\frac{1 - \\left(
            \\frac{(1 + R_{\\text{in}}) \\cdot \\sigma_{amp, \\text{in}}}
            {(1 - R_{\\text{in}}) \\cdot \\sigma_{ult}}
            \\right)^{n}}{\\sigma_{amp, \\text{in}}}
            + \\sigma_{amp, \\text{out}}^{n-1} \\cdot \\left(
            \\frac{(1 + R_{\\text{out}})}{(1 - R_{\\text{out}})
            \\cdot \\sigma_{ult}} \\right)^{n}}

    where:

    * :math:`R`\\ :sub:`in`\\  = :math:`\\sigma_{min} / \\sigma_{max} =
      (\\sigma_{mean} - \\sigma_{amp}) / (\\sigma_{mean} + \\sigma_{amp})` is
      the input stress ratio,
    * :math:`R`\\ :sub:`out`\\  is the output stress ratio,
    * :math:`\\sigma_{amp}`\\ :sub:`in`\\  is the input amplitude,
    * :math:`\\sigma_{amp}`\\ :sub:`out`\\  is the output amplitude,
    * :math:`\\sigma_{ult}` is the ultimate strength of the material, and
    * :math:`n` is the correction exponent.

    The solver is based on the Newton-Raphson method, which is implemented
    internally at :func:`py_fatigue.utils.numba_newton`,
    :func:`py_fatigue.utils.compile_specialized_newton`, and
    :func:`py_fatigue.mean_stress.corrections.__goodman_equation`.

    Parameters
    ----------
    amp_in : np.ndarray | list
        Input amplitude values.
    mean_in : np.ndarray | list
        Input mean stress values.
    r_out : float | np.ndarray
        Output stress ratio values.
    ult_s : float
        Ultimate strength of the material.
    correction_exponent : float
        Mean stress correction exponent (n).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Computed amp_out values, and corresponding mean_out values.

    See Also
    --------
    :func:`py_fatigue.utils.numba_newton`,
    :func:`py_fatigue.utils.compile_specialized_newton`,
    :func:`py_fatigue.mean_stress.corrections.__goodman

    Raises
    ------
    ValueError
        If the input arrays have different shapes or if there are negative
        stress amplitude values in amp_in.
    Warning
        If the number of points in the input arrays is too large for plotting.

    Examples
    --------
    >>> import numpy as np
    >>> from py_fatigue.mean_stress.corrections import (
    ...     goodman_haigh_mean_stress_correction
    ... )
    >>> amp_in = np.array([100, 200, 300])
    >>> mean_in = np.array([0, 0, 0])
    >>> r_out = np.array([-1, 0, 0.5, 0.75, 0.9])
    >>> ult_s = 1000
    >>> correction_exponent = 9
    >>> amp_out, mean_out = goodman_haigh_mean_stress_correction(
    ...     amp_in, mean_in, r_out, ult_s, correction_exponent, plot=True
    ... )

    """
    # Convert inputs to numpy arrays for element-wise operations
    if isinstance(amp_in, list):
        amp_in = np.asarray(amp_in, dtype=np.float64)
    if isinstance(mean_in, list):
        mean_in = np.asarray(mean_in, dtype=np.float64)
    if np.isscalar(r_out):
        r_out = np.array([r_out], dtype=np.float64)
    r_out = np.asarray(r_out, dtype=np.float64)
    # Check that the input arrays have the same shape
    if amp_in.shape != mean_in.shape:
        raise ValueError("amp_in and mean_in must have the same shape")
    # Check that all values in amp_in are positive
    if np.any(amp_in < 0):
        raise ValueError("You have negative stress amplitude values in amp_in")
    # Check that, if the product of the size of amp_in and r_out is greater
    # than 1,000,000, the user has set plot to False
    if amp_in.size * r_out.size > 1_000_000 and plot:
        if logger is not None:
            # fmt: off
            logger.warning("\033[1mThe number of points in the input arrays "
                           "is too large for plotting\033[0m")
            logger.info("Setting plot to False to avoid matplotlib errors")
        warnings.warn(
            "The number of points in the input arrays is too large "
            "for plotting. Setting plot to False to avoid "
            "matplotlib errors"
        )

        # fmt: on
        plot = False
    # Compute r_in from amp_in and mean_in
    r_in = (mean_in - amp_in) / (mean_in + amp_in)
    # mean_in = amp_in * (1 + r_in) / (1 - r_in)
    # Initial guess for amp_out set to Goodman Correction for r_out = - 1
    if initial_guess is None:
        initial_guess = amp_in / (1 - (mean_in / ult_s) ** correction_exponent)
    # Set initial guess equal to zero wherever r_in > 1
    initial_guess[r_in > 1] = 0

    # Solve for each amp_out using fsolve
    amp_out = np.zeros((len(r_out), len(initial_guess)))
    mean_out = np.zeros((len(r_out), len(initial_guess)))
    # amp_out = []
    # mean_out = []
    for r_out_val in r_out:
        # NOTE: Special cases for r_out = -1 and r_out = 0
        # NOTE: These cases are solved analytically to improve performance
        if r_out_val == -1:
            # amp_out.append(
            #     amp_in / (1 - (mean_in / ult_s) ** correction_exponent)
            # )
            # mean_out.append(np.zeros_like(amp_in))
            amp_out[r_out == -1, :] = amp_in / (
                1 - (mean_in / ult_s) ** correction_exponent
            )
            mean_out[r_out == -1, :] = np.zeros_like(amp_in)
            continue
        amp_out_fsolve = []
        for i in range(len(initial_guess)):
            # Solve the implicit equation using fsolve
            # fmt: off
            if mean_in[i] + amp_in[i] > 0:
                sol = numba_newton(__jit_goodman_equation, initial_guess[i],
                                    1E-6, 1000, amp_in[i], r_in[i],
                                    r_out_val, ult_s, correction_exponent)
                # sol = fsolve(__goodman_equation, x0=initial_guess[i],
                #              args=(amp_in[i], r_in[i], r_out_val, ult_s,
                #                    correction_exponent), xtol=1E-6)
            else:
                sol = amp_in[i]
            # Check if fsolve converged or not
            if sol is None:
                if logger is not None:
                    logger.warning(f"fsolve did not converge for amp_in="
                                   f"{amp_in[i]}, r_in={r_in[i]}, r_out="
                                   f"{r_out_val}")
                amp_out_fsolve.append(np.nan)
            else:
                amp_out_fsolve.append(sol if sol < ult_s else ult_s)
            # fmt: on
        # amp_out.append(amp_out_fsolve)
        # mean_out.append(
        #     np.array(amp_out_fsolve) * (1 + r_out_val) / (1 - r_out_val)
        # )
        amp_out[r_out == r_out_val, :] = np.asarray(amp_out_fsolve)
        mean_out[r_out == r_out_val, :] = (
            amp_out[r_out == r_out_val, :] * (1 + r_out_val) / (1 - r_out_val)
        )

    # Sort r_out, as well as amp_out and mean_out (by r_out)
    if np.all(r_out[:-1] <= r_out[1:]):
        srt_r_out = r_out.copy()
        srt_amp_out = amp_out.copy()
        srt_mean_out = mean_out.copy()
    else:
        srt_idx = np.argsort(r_out)
        srt_r_out = r_out[srt_idx]
        srt_amp_out = amp_out[srt_idx, :]
        srt_mean_out = mean_out[srt_idx, :]
    if plot:
        # Create figure with two subplots
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))

        # First subplot: amp_out vs r_out
        # fmt: off
        for i in range(len(amp_in)):
            ax1.plot(srt_r_out, srt_amp_out[:, i],
                    marker='.', markersize=2, linewidth=0.5,
                    label=f"$amp_{{in}}={np.round(amp_in[i], 1)}, "
                        f"r_{{in}}={np.round(r_in[i], 2)}$")
        ax1.plot(r_in, amp_in,
                'o', color='yellow', markersize=4, markeredgecolor='black',
                markeredgewidth=0.5, label='Original Data')
        ax1.set_xlabel("load ratio, $\\sigma_{min}/ \\sigma_{max}$")
        ax1.set_ylabel("MSC stress amplitude, $\\Delta \\sigma_{MSC} / 2$")
        ax1.set_title("Goodman Correction - Amplitude vs R")
        ax1.set_xlim(right=1.25, left=max(-5, min(r_in) - 0.25))
        ax1.set_ylim(top=min(max(amp_in) * 1.5, ult_s))
        ax1.grid(True)

        # Second subplot: mean_out vs amp_out
        for i in range(len(amp_in)):
            ax2.plot(srt_mean_out[:, i], srt_amp_out[:, i],
                    marker='.', markersize=2, linewidth=0.5,
                    label=f"$amp_{{in}}={np.round(amp_in[i], 1)}, "
                        f"r_{{in}}={np.round(r_in[i], 2)}$")
        # stress_amplitude * (1 + load_ratio) / (1 - load_ratio)
        ax2.plot(amp_in * (1 + r_in) / (1 - r_in), amp_in,
                'o', color='yellow', markersize=4, markeredgecolor='black',
                markeredgewidth=0.5, label='Original Data')

        # Set limits to maximum stress amplitude and ultimate strength
        ax2.set_xlim(right=ult_s * 1.1)
        ax2.set_ylim(top=min(max(amp_in) * 1.5, ult_s))
        # draw lines at constant load ratio for better visualization

        for r in [-0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75]:
            sample_mean_out = np.linspace(0, ult_s * i, 2)
            ax2.plot(sample_mean_out, (1 - r) / (1 + r) * sample_mean_out, ':',
                    color='#AAAAAA', linewidth=0.75)
        ax2.set_xlabel("MSC mean stress, $\\sigma_{m,MSC}$")
        ax2.set_ylabel("MSC stress amplitude, $\\Delta \\sigma_{MSC} / 2$")
        ax2.set_title("Goodman Correction - Amplitude vs Mean")
        handles, labels = ax2.get_legend_handles_labels()
        ax2.grid(False)

        # Adjust layout and show legend
        plt.tight_layout()
        if len(handles) < 31:
            fig.legend(handles, labels, bbox_to_anchor=(0.5, -0.05),
                       loc='upper center', ncol=4)
        # fmt: on
        plt.show()
    return amp_out, mean_out


def __goodman_equation(
    amp_out_val: float,
    amp_in_val: float,
    r_in_val: float,
    r_out_val: float,
    ult_s: float,
    correction_exponent: float,
) -> float:
    """Implicit equation to solve for amp_out.

    .. math::

        \\sigma_{amp, \\text{out}} =  \\frac{1}{ \\frac{1 - \\left(
            \\frac{(1 + R_{\\text{in}}) \\cdot \\sigma_{amp, \\text{in}}}
            {(1 - R_{\\text{in}}) \\cdot \\sigma_{ult}}
            \\right)^{n}}{\\sigma_{amp, \\text{in}}}
            + \\sigma_{amp, \\text{out}}^{n-1} \\cdot \\left(
            \\frac{(1 + R_{\\text{out}})}{(1 - R_{\\text{out}})
            \\cdot \\sigma_{ult}} \\right)^{n}}
    where:
    * :math:`R`\\ :sub:`in`\\  = :math:`\\sigma_{min} / \\sigma_{max} =
      (\\sigma_{mean} - \\sigma_{amp}) / (\\sigma_{mean} + \\sigma_{amp})` is
      the input stress ratio,
    * :math:`R`\\ :sub:`out`\\  is the output stress ratio,
    * :math:`\\sigma_{amp}`\\ :sub:`in`\\  is the input amplitude,
    * :math:`\\sigma_{amp}`\\ :sub:`out`\\  is the output amplitude,
    * :math:`\\sigma_{ult}` is the ultimate strength of the material, and
    * :math:`n` is the correction exponent.
    Parameters
    ----------
    amp_out_val : float
        The output amplitude value.
    amp_in_val : float
        The input amplitude value.
    r_in_val : float
        The input load ratio value.
    r_out_val : float
        The output load ratio value.
    ult_s : float
        The ultimate strength of the material.
    correction_exponent : float
        The correction exponent value.
    Returns
    -------
    float
        The difference between the left-hand side and right-hand side of the
        equation.
    """
    assert r_in_val != 1, "Input load ratio equals 1, meaning static load."
    # if r_in_val == 1:
    # fmt: off
    # raise ValueError(f"Input load ratio equals 1 at amp_in={amp_in_val}"
    #                   ", meaning static load. Revise input data.")
    # fmt: on
    trm1 = (1 + r_in_val) / (1 - r_in_val) * amp_in_val / ult_s
    trm2 = (1 + r_out_val) / (1 - r_out_val) / ult_s

    lhs = amp_out_val
    # fmt: off
    rhs = (1 - trm1**correction_exponent) / amp_in_val + \
            amp_out_val**(correction_exponent - 1) * trm2**correction_exponent
    # fmt: on
    return lhs - rhs**-1


__jit_goodman_equation = compile_specialized_newton(__goodman_equation)



# -*- coding: utf-8 -*-
"""The module :mod:`py_fatigue.damage.stress_life` contains all the
damage models related to the stress-life approach.
"""

# Packages from standard library

from __future__ import annotations
from types import SimpleNamespace
from typing import Any, Callable, Optional, Tuple, Union
import logging
import warnings

# Packages from external libraries
import matplotlib
import matplotlib.axes
import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import pandas as pd

# py-fatigue imports
from ..cycle_count.cycle_count import CycleCount
from ..material.sn_curve import (
    _check_param_couple_types,
    __jit_sn_curve_residuals,
    SNCurve,
)
from ..styling import py_fatigue_formatwarning
from ..utils import make_axes, numba_bisect, _plot_damage_accumulation

try:
    # delete the accessor to avoid warning
    del pd.DataFrame.miner  # type: ignore
except AttributeError:
    pass


warnings.formatwarning = py_fatigue_formatwarning


@pd.api.extensions.register_dataframe_accessor("miner")
class PalmgrenMiner:
    """Accessor for the Palmgren-Miner damage model."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self.sn_curve = None
        self.des_data = {}
        self.dem_data = {}

    @staticmethod
    def _validate(obj):
        """Validate the input DataFrame. Raise an error if the input
        DataFrame does not contain the right columns.
        """
        if {
            "cycles_to_failure",
        }.issubset(obj.columns):
            e_msg = "'cycles_to_failure' already calculated"
            raise AttributeError(e_msg)

        if not {"count_cycle", "mean_stress", "stress_range"}.issubset(
            obj.columns
        ):
            e_msg = "Must have 'count_cycle', 'mean_stress' and 'stress_range'."
            raise AttributeError(e_msg)

    def damage(self, sn_curve: SNCurve):
        """Calculate the damage of the samples.

        Parameters
        ----------
        sn_curve : SNCurve
            The SNCurve object.

        Returns
        -------
        DataFrame
            The DataFrame with the damage.
        """
        if self._obj._metadata["unit"] != sn_curve.unit:
            e_msg = (
                f"Units of cycle_count ({self._obj._metadata['unit']}) and "
                f"sn_curve ({sn_curve.unit}) do not match."
            )
            raise ValueError(e_msg)
        self._obj.sn_curve = sn_curve
        self.sn_curve = sn_curve
        self._obj["cycles_to_failure"] = sn_curve.get_cycles(
            self._obj.stress_range
        )
        self._obj["pm_damage"] = np.divide(
            self._obj.count_cycle, self._obj.cycles_to_failure
        )
        return self._obj

    def des(self, slope: float, equivalent_cycles: float = 1e7) -> float:
        """Calculate the damage equivalent stress range (DES).

        Parameters
        ----------
        slope : float
            The SN curve slope.
        equivalent_cycles : float, optional
            The equivalent number of cycles, by default 1e7

        Returns
        -------
        float
            The DES.

        See Also
        --------
        calc_des
        """
        self.des_data["slope"] = slope
        self.des_data["equivalent_cycles"] = equivalent_cycles

        return calc_des(
            self._obj.stress_range,
            self._obj.count_cycle,
            slope,
            equivalent_cycles,
        )

    def dem(
        self,
        outer_radius: float,
        inner_radius: float,
        slope: float,
        equivalent_cycles: float = 1e7,
    ) -> float:
        """Calculate the damage equivalent stress range (DES).

        Parameters
        ----------
        slope : float
            The SN curve slope.
        equivalent_cycles : float, optional
            The equivalent number of cycles, by default 1e7

        Returns
        -------
        float
            The DES.

        See Also
        --------
        calc_des
        """
        self.dem_data["outer_radius"] = outer_radius
        self.dem_data["inner_radius"] = inner_radius
        self.dem_data["slope"] = slope
        self.dem_data["equivalent_cycles"] = equivalent_cycles

        return calc_dem(
            outer_radius,
            inner_radius,
            self._obj.stress_range,
            self._obj.count_cycle,
            slope,
            equivalent_cycles,
        )

    def plot_histogram(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        dens_func: Optional[Callable] = None,
        **kwargs,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """Plot the damage.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure to plot on, by default None
        ax : matplotlib.collections.PathCollection, optional
            The axes to plot on, by default None
        kwargs
            Additional keyword arguments for the plot.

        Returns
        -------
        matplotlib.figure.Figure, matplotlib.collections.PathCollection
            The figure and axes.
        """
        fig, ax = make_axes(fig, ax)
        im = ax.scatter(
            self._obj.count_cycle,
            self._obj.stress_range,
            # edgecolors="#CCC",
            # linewidths=0.5,
            c=(
                dens_func(self._obj.pm_damage)
                if dens_func is not None
                else self._obj.pm_damage
            ),
            **kwargs,
        )
        cbar_label = (
            "PM damage"
            if dens_func is None
            else "PM damage(" + str(dens_func) + ")"
        )
        ax.set_xlabel("# of cycles")
        ax.set_ylabel("Stress range, MPa")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label(cbar_label, rotation=270)
        cbar.ax.get_yaxis().labelpad = 10
        return fig, ax


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.gassner  # type: ignore
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("gassner")
class Gassner:
    """Accessor for the Gassner shift factor for variable amplitude."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self.sn_curve = None

    @staticmethod
    def _validate(obj):
        """Validate the input DataFrame. Raise an error if the input
        DataFrame does not contain the right columns.
        """
        if {
            "shift_factor",
        }.issubset(obj.columns):
            e_msg = "'shift_factor' already calculated"
            raise AttributeError(e_msg)

        if not {"count_cycle", "mean_stress", "stress_range"}.issubset(
            obj.columns
        ):
            e_msg = "Must have 'count_cycle', 'mean_stress' and 'stress_range'."
            raise AttributeError(e_msg)

    def g(self, sn_curve: SNCurve):
        """Calculate the shift factor of the samples.

        The shift factor is a spectrum-dependent factor that allows
        to shift a Wohler curve and avoids having to calculate the
        damage using the :class:`~PalmgrenMiner` ruleevery time.

        Once the shift factor has been calculated, it can be applied
        to the Wohler SN curve. Then, accssing the curve through the
        maximum amplitude value of the spectrum will return the
        expected life for said spectrum.

        The shift factor for each cycle  in the spectrum is:

        .. math::

            G_j = \\frac{n_j}{N_S} \\,
            \\left({\\frac{\\sigma_{{max}_j} \\, \\sigma_{{alt}_j}}
            {\\max{\\sigma_{alt}} ^ {2}}}\\right)^{2}

        The shift factor then is:

        .. math::

            G = \\sum_{j=1}^{N_B} G_j

        Parameters
        ----------
        sn_curve : SNCurve
            The SNCurve object.

        Returns
        -------
        DataFrame
            The DataFrame with the shift factor per sample.
        """
        assert sn_curve.linear, "The SN curve must be linear"
        self._obj.sn_curve = sn_curve
        self.sn_curve = sn_curve
        stress_amp = self._obj.stress_range / 2
        max_amp = max(stress_amp)
        nu = np.divide(self._obj.count_cycle, sum(self._obj.count_cycle))
        self._obj["shift_factor"] = nu * np.power(
            np.divide(stress_amp, max_amp), self.sn_curve.slope
        )
        return self._obj


def calc_pm(
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    sn_curve: SNCurve,
) -> np.ndarray:
    """Calculates the damage using the Palmgren-Miner rule, i.e. the
    linear damage accumulation rule:

    .. math::

        D = \\sum_{j=1}^{N_{\\text{blocks}}} \\frac{n_j}{N_j} \\leq 1

    Parameters
    ----------
     stress_range : Union[pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[pd.Series, np.ndarray]
        The number of cycles
    sn_curve : SNCurve
        The SNCurve object.

    Returns
    -------
    Union[float, np.ndarray]
        The damage according to the Palmgren-Miner rule.
    """
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    return count_cycle / sn_curve.get_cycles(stress_range)


def calc_des(
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent stress range (DES).

    Parameters
    ----------
    stress_range : Union[float, pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[float, pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DES.
    """
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    des = np.power(
        np.sum(
            np.multiply(
                np.power(stress_range, float(slope)),
                count_cycle,
            )
            / equivalent_cycles
        ),
        1 / slope,
    )
    return des


def calc_dem(
    outer_radius: float,
    inner_radius: float,
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent moment (DEM) in Nm.

    .. math::

        DEM = DES * I_c / r_o = DES * S

    with :math:`I_c` the area moment of inertia and :math:`r_i` the
    inner radius, or :math:`S` the Section modulus. The Damage
    Equivalent Stress Range (DES) is calculated using
    :func:`calc_des`

    Parameters
    ----------
    outer_radius : float
        The outer radius to calculate the DEM. Must be provided in m
    inner_radius : float
        The inner radius to calculate the DEM. Must be provided in m
    stress_range : Union[float, pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[float, pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Raises
    ------
    ValueError
        If the inner radius is larger than the outer radius.

    Returns
    -------
    float
        The DEM in Nm.
    """
    if outer_radius <= inner_radius:
        raise ValueError("outer_radius must be greater than inner_radius")
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    area_inertia = np.pi / 4 * (outer_radius**4 - inner_radius**4)
    des = calc_des(stress_range, count_cycle, slope, equivalent_cycles)

    return area_inertia * des * 1e6 / outer_radius


def get_pm(cycle_count: CycleCount, sn_curve: SNCurve) -> np.ndarray:
    """Calculates the damage using the Palmgren-Miner rule.

    Parameters
    ----------
    cycle_count : CycleCount
        The CycleCount object.
    sn_curve : SNCurve
        The SNCurve object.

    Returns
    -------
    np.ndarray
        The damage according to the Palmgren-Miner rule.
    """
    if cycle_count.unit != sn_curve.unit:
        raise ValueError(
            f"Units of cycle_count ({cycle_count.unit}) and "
            f"sn_curve ({sn_curve.unit}) do not match."
        )
    return calc_pm(cycle_count.stress_range, cycle_count.count_cycle, sn_curve)


def get_des(
    cycle_count: CycleCount, slope: float, equivalent_cycles: float = 1e7
) -> float:
    """Calculate the damage equivalent stress range (DES).

    Parameters
    ----------
    cycle_count : CycleCount
        The CycleCount object.
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DES.
    """
    return calc_des(
        cycle_count.stress_range,
        cycle_count.count_cycle,
        slope,
        equivalent_cycles,
    )


def get_dem(
    outer_radius: float,
    inner_radius: float,
    cycle_count: CycleCount,
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent moment (DEM) in Nm.

    .. math::

        DEM = DES * I_c / r_o = DES * S

    with :math:`I_c` the area moment of inertia and :math:`r_i` the
    inner radius, or :math:`S` the Section modulus. The Damage
    Equivalent Stress Range (DES) is calculated using
    :func:`calc_des`

    Parameters
    ----------
    outer_radius : float
        The outer radius to calculate the DEM. Must be provided in m
    inner_radius : float
        The inner radius to calculate the DEM. Must be provided in m
    stress_range : Union[pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DEM in Nm.
    """
    return calc_dem(
        outer_radius,
        inner_radius,
        cycle_count.stress_range,
        cycle_count.count_cycle,
        slope,
        equivalent_cycles,
    )


def calc_manson_halford_exponents(
    cycles_to_failure: np.ndarray, exponent: float = 0.4
) -> np.ndarray:
    """Calculate the Mannson-Halford exponents :math:`e_{j, j+1}`.

    .. math::

        e{j, j+1} = \\left(\\frac{N_{j}}{N_{j+1}}\\right)^{exponent}

    Parameters
    ----------
    cycles_to_failure : np.ndarray
        The number of cycles to failure.
    exponent : float, optional
        The exponent, by default 0.4

    Returns
    -------
    np.ndarray
        The Mannson-Halford exponents.
    """
    return (
        np.hstack([cycles_to_failure[:-1] / cycles_to_failure[1:], [1]])
        ** exponent
    )


def calc_pavlou_exponents(
    stress_range: np.ndarray,
    ultimate_stress: float = 900,
    exponent: float = -0.75,
    use_dca: bool = False,
) -> np.ndarray:
    """Calculate the Pavlou exponents :math:`q(\\sigma_j)=e_{j, j+1}`.

    .. math::

        e_{j, j+1} = \\left(\\frac{\\Delta \\sigma_j / 2}{\\sigma_U}\\right)
        ^{exponent}

    where :math:`\\Delta \\sigma_j` is the stress amplitude, :math:`\\sigma_U`
    is the ultimate stress, :math:`\\Delta \\sigma` is the stress range and
    :math:`exponent` is the exponent.


    Parameters
    ----------
    stress_range : np.ndarray
        The stress range.
    ultimate_stress : float, optional
        The ultimate stress, by default 900
    exponent : float, optional
        The exponent, by default -0.75

    Returns
    -------
    np.ndarray
        The Pavlou exponents.
    """
    q_exp = (stress_range / 2 / ultimate_stress) ** exponent
    if use_dca:
        return q_exp
    return np.hstack([q_exp[1:] / q_exp[:-1], [1]])


def calc_si_jian_exponents(
    stress_range: np.ndarray,
) -> np.ndarray:
    """Calculate the Si-Jian et al. exponents as

    .. math::

        e_{j, j+1} = \\sigma_{j+1} / \\sigma_{j}

    where :math:`\\sigma_{j+1}` is the stress amplitude for the
    :math:`j`-th cycle.

    Parameters
    ----------
    stress_range : np.ndarray
        The stress range.

    Returns
    -------
    np.ndarray
        The Si-Jian et al. exponents.
    """
    return np.hstack([stress_range[1:] / stress_range[:-1], [1]])


def calc_theil_weights(
    stress_range: np.ndarray,
    sn_curve: SNCurve,
) -> np.ndarray:
    """Calculate the Theil weights for the Damage Curve Approach
    (DCA).

    .. math::

        w_j = \\frac{\\Delta\\sigma_j}{N_{f,j}}

    Parameters
    ----------
    stress_range : np.ndarray
        The stress range.
    sn_curve : SNCurve
        The py-fatigue SN Curve

    Returns
    -------
    np.ndarray
        Theil's model weights
    """
    return np.divide(stress_range, sn_curve.get_cycles(stress_range))


def _calc_damage_exponents(
    damage_rule: str,
    stress_range: np.ndarray,
    sn_curve: Optional[SNCurve] = None,
    **kwargs: dict,
):
    """Calculate damage exponents based on the specified damage rule.
    Parameters
    ----------
    damage_rule : str
        The name of the damage rule to use. Supported rules are:
        'pavlou', 'manson', 'leve', and 'si jian'.
    stress_range : array_like
        An array-like object containing the stress ranges.
    sn_curve : SNCurve, optional
        An object representing the S-N curve.  It must have a `get_cycles`
        method that accepts a stress range and returns the number of cycles
        to failure. This is only used for the 'manson' damage rule.
    **kwargs : dict, optional
        Additional keyword arguments required by specific damage rules.
        - For 'pavlou':
            - 'base_exponent' (float): The base exponent for the Pavlou
              damage rule. If not provided, a default value of -0.75 is
              used.
            - 'ultimate_stress' (float): The ultimate stress for the Pavlou
              damage rule. If not provided, a default value of 900 MPa is
              used.
        - For 'manson':
            - 'base_exponent' (float): The base exponent for the
              Manson-Halford damage rule. If not provided, a default value
              of 0.4 is used.
        - For 'leve':
            - 'base_exponent' (float): The base exponent for the Leve
              damage rule. If not provided, a default value of 2 is used.
    Returns
    -------
    numpy.ndarray
        An array of damage exponents calculated based on the specified
        damage rule.
    Raises
    ------
    ValueError
        If an unknown damage rule is specified.
    UserWarning
        If required keyword arguments are missing for a specific damage
        rule, a warning is issued and a default value is used.
        The warnings are formatted using `py_fatigue_formatwarning`.
    Notes
    -----
    The 'pavlou' damage rule requires 'base_exponent' and 'ultimate_stress'
    to be specified in kwargs. If not specified, default values are used
    and a warning is issued.
    The 'manson' damage rule requires 'base_exponent' to be specified in
    kwargs. If not specified, a default value is used and a warning is
    issued. The S-N curve object must have a `get_cycles` method.
    The 'leve' damage rule requires 'base_exponent' to be specified in
    kwargs. If not specified, a default value is used and a warning is
    issued.
    The 'si jian' damage rule does not require any additional keyword
    arguments.
    """
    ns = SimpleNamespace(**kwargs)
    if damage_rule.lower() == "pavlou":
        if "base_exponent" not in kwargs:
            w_msg = [
                "Pavlou damage rule requires 'base_exponent' ",
                "to be assigned.\nUsing preset value of -0.75.",
            ]
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = -0.75
        if "ultimate_stress" not in kwargs:
            w_msg = [
                "Pavlou damage rule requires 'ultimate_stress' ",
                "to be assigned.\nUsing preset value of 900 MPa.",
            ]
            warnings.warn("".join(w_msg), UserWarning)
            ns.ultimate_stress = 900
        if "use_dca" not in kwargs:
            w_msg = [
                "Pavlou damage rule requires 'use_dca' (Damage Curve Approach)"
                "to be assigned.\nUsing preset value of true.",
            ]
            warnings.warn(" ".join(w_msg), UserWarning)
            ns.use_dca = False
        return calc_pavlou_exponents(
            stress_range, ns.ultimate_stress, ns.base_exponent, ns.use_dca
        )
    if "manson" in damage_rule.lower():
        if "base_exponent" not in kwargs:
            w_msg = [
                "Manson & Halford damage rule requires ",
                "'base_exponent' to be assigned.\n",
                "Using preset value of 0.4.",
            ]
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = 0.4
            if not sn_curve:
                raise ValueError("sn_curve must be provided for 'manson'")
        return calc_manson_halford_exponents(
            sn_curve.get_cycles(stress_range), ns.base_exponent  # type: ignore
        )
    if damage_rule.lower() == "leve":
        if "base_exponent" not in kwargs:
            w_msg = [
                "Leve damage rule requires ",
                "'base_exponent' to be assigned.\n",
                "Using preset value of 2.",
            ]
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = 2
        return ns.base_exponent * np.ones(len(stress_range))
    if "si jian" in damage_rule.lower():
        return calc_si_jian_exponents(stress_range)

    raise ValueError(f"Unknown damage rule: {damage_rule}")


def calc_nonlinear_damage(
    damage_rule: str,
    stress_range: np.ndarray,
    count_cycle: np.ndarray,
    sn_curve: SNCurve,
    **kwargs,
) -> float:
    """Calculate the fatigue damage using a nonlinear damage rule
    among the allowed ones:

    - 'Pavlou': Pavlou damage rule
    - 'Manson-Halford': Mannson-Halford damage rule
    - 'Si-Jian': Si-Jian et al damage rule
    - 'Leve': Leve damage rule

    .. warning::
        This function is not suited for variable amplitude loading.
        For variable amplitude loading, use
        :func:`calc_nonlinear_damage_with_dca`, as the fluctiations of the
        nonlinear damage exponent are controlled by the Damage Curve.

    The generic form of a nonlinear damage rule is:

    .. math::

        D = \\left(
            \\left( \\dots
                \\left(
                    \\left(
                        \\left(\\frac{n_1}{N_1}\\right)^{e_{1, 2}} +
                        \\frac{n_2}{N_2}
                    \\right)^{e_{2, 3}} +
                    \\frac{n_3}{N_3}
                \\right)^{e_{3, 4}} + \\dots + \\frac{n_{M-1}}{N_{M-1}}
            \\right)^{e_{M-1, M}} + \\dots + \\frac{n_M}{N_M}
        \\right)^{e_M}

    where :math:`n_j` is the number of cycles in the fatigue histogram
    at the :math:`j`-th cycle, :math:`N_j` is the number of cycles to
    failure at the :math:`j`-th cycle, :math:`e_{j, j+1}` is the exponent
    for the :math:`j`-th and :math:`j+1`-th cycles, :math:`M` is the
    number of load blocks in the fatigue spectrum.

    The formula is conveniently rewritten as pseudocode:

    .. code-block:: python
        :caption: pseudocode for the nonlinear damage rule

        # retrieve N_j using the fatigue histogram and SN curve
        # retrieve the exponents e_{j, j+1}
        #  calculate the damage
        D = 0
        for j in range(1, M+1):
            D = (D + n_j / N_j) ^ e_{j, j+1}

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Manson-Halford', 'Si-Jian', 'Leve'.
    stress_range : np.ndarray
        The stress range.
    count_cycle : np.ndarray
        The number of cycles.
    sn_curve : SNCurve
        The SN curve.
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    np.ndarray
        The cumulated damage.
    """
    damage_per_cycle = calc_pm(
        stress_range,
        count_cycle,
        sn_curve,
    )

    kwargs["use_dca"] = False
    damage_exp = _calc_damage_exponents(
        damage_rule, stress_range, sn_curve, **kwargs
    )

    total_damage = 0
    damage_array = np.empty(len(damage_per_cycle))
    for i, (d_i, e_i) in enumerate(zip(damage_per_cycle, damage_exp)):
        # total_damage = (total_damage + d_i) ** e_i - total_damage
        total_damage = (total_damage + d_i) ** e_i
        damage_array[i] = total_damage

    return damage_array  # type: ignore


def calc_nonlinear_damage_with_dca(
    damage_rule: str,
    stress_range: np.ndarray,
    count_cycle: np.ndarray,
    sn_curve: SNCurve,
    damage_bands: np.ndarray = np.array(
        [0, 0.025, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1]
    ),
    limit_damage: Union[float, int] = 1,
    logger: logging.Logger | None = None,
    plot: bool = False,
    **kwargs,
) -> tuple[
    np.ndarray,
    np.ndarray,
    matplotlib.figure.Figure | None,
    matplotlib.axes.Axes | None,
]:
    """Calculate the fatigue damage using a nonlinear damage rule based on the
    Damage Curve Approach (DCA) among the allowed ones:

    - 'Pavlou': Pavlou damage rule
    - 'Theil': Theil damage rule

    The DCA discreties the damage curve in multiple Damage bands
    :math:`\\Delta D_j` and calculates the damage for each band, depending
    on the damage value at the previous cycle. The damage is calculated
    as:

    .. math::

        D = \\sum_{j=1}^{n_b} \\Delta D_j

    where :math:`n_b` is the number of damage bands. In each damage band,
    the damage follows a weighted Palmgren-Miner sum, i.e.:

    .. math::

        \\Delta D_j = \\sum_{i=1}^{n_j} w_{i, j} \\frac{n_i}{N_i}

    where :math:`n_j` is the number of cycles in the fatigue histogram
    at the :math:`j`-th cycle, :math:`N_j` is the number of cycles to
    failure at the :math:`j`-th cycle, :math:`w_{i, j}` is the weight
    for the :math:`i`-th cycle in the :math:`j`-th damage band.

    Using the generic nonlinear damage accumulation formula:

    .. math::

        D(\\sigma) = \\left(\\frac{n}{N(\\sigma)}\\right)^{e(\\sigma)}

    and substituting it inside the weighted Palmgren-Miner sum, the
    equation of the weights can be extracted as:

    .. math::

        w_{i, j} = \\frac{D_j - D_{j-1}}{
            D_{j}^{1/e_{i, i}} - D_{j-1}^{1/e_{i, i}}
        }

    In the case of Pavlou's model, the exponents are:

    .. math::

        e_{j, j} = \\left(\\frac{\\Delta \\sigma_j / 2}
                                {\\sigma_U}\\right)^{exponent}

    While for Theil's method, the weights can be directly
    approximated as:

    .. math::

        w_j = 1 + \\frac{\\Delta\\sigma_j}{N_{f,j}}

    .. warning::
        If you want to use Theil's method in a production
        environment, it is recommended to use the
        :func:`calc_theil_cycles_to_failure` function, as it adheres
        to the original SN Cuve-based method.

    The formula is conveniently rewritten as pseudocode:

    .. code-block:: python
        :caption: pseudocode for the nonlinear damage rule

        # retrieve n_i, N_i using the fatigue histogram and SN curve
        # retrieve the exponents e_{j, j+1}
        # define the damage bands DB
        # calculate the damage
        D = 0
        for j in range(1, M+1):
            DB_j = np.digitize(D, DB, right=False)
            for i in range(1, n_j+1):
                if damage_rule == 'Pavlou':
                    w_ij = (DB[DB_j] - DB[DB_j - 1]) / \
                           ((DB[DB_j] ** (1 / e_i)) -
                            (DB[DB_j - 1] ** (1 / e_i)))
                elif damage_rule == 'Theil':
                    w_ij = stress_range_i / N_i + 1
                D += w_ij * n_i / N_i

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Theil'.
    stress_range : np.ndarray
        The stress range.
    count_cycle : np.ndarray
        The number of cycles.
    sn_curve : SNCurve
        The SN curve.
    damage_bands : np.ndarray
        Damage bands for the Pavlou damage calculation, by default
        [0, 0.025, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1]
    limit_damage : float, optional
        Limit damage value, by default 1
    logger: logging.Logger, optional
        If a logger is defined, attach output to it
    plot: bool, by default False
        Whether to plot or not the cumulated damage
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, mpl.Figure | None, mpl.Axes | None]
        The cycle-to-cycle damage and the cumulated damage.
    """

    allowed_damage_rules = ["pavlou", "theil"]
    if damage_rule.lower() not in allowed_damage_rules:
        raise ValueError(
            f"Unknown damage rule: {damage_rule}. "
            f"Allowed damage rules are: {allowed_damage_rules}"
        )
    stress_range = np.asarray(stress_range, dtype=np.float64)
    count_cycle = np.asarray(count_cycle, dtype=np.float64)
    damage_bands = np.asarray(damage_bands, dtype=np.float64)
    pm_damage = calc_pm(
        stress_range,
        count_cycle,
        sn_curve,
    )

    if damage_rule.lower() == "theil":
        weights = calc_theil_weights(stress_range, sn_curve)
        nl_damage = (1 + weights) * pm_damage
        if not plot:
            return nl_damage, np.cumsum(nl_damage), None, None
        fig, ax = _plot_damage_accumulation(
            np.cumsum(nl_damage), np.cumsum(pm_damage), limit_damage
        )
        return nl_damage, np.cumsum(nl_damage), fig, ax

    kwargs["use_dca"] = True
    dmg_exp = _calc_damage_exponents(
        damage_rule, stress_range, sn_curve, **kwargs
    )

    nl_damage = pm_damage.copy()
    cumsum_nl_dmg = np.zeros_like(nl_damage)
    cur_dmg_band = 0
    for i in range(len(pm_damage)):  # pylint: disable=C0200
        prev_dmg_band = cur_dmg_band if i >= 1 else 0
        # fmt: off
        cur_dmg_band = np.digitize(cumsum_nl_dmg[i - 1], damage_bands,
                                   right=False) if i >= 1 else 0
        cur_dmg_band = min(cur_dmg_band, len(damage_bands) - 1)
        if (
            cur_dmg_band != prev_dmg_band
            and len(damage_bands) < 20
            and logger is not None
        ):
            if cur_dmg_band == 1:
                logger.info("Damage band change")
            logger.info(f"‣ at cycle {i}:\n"
                        f"                • from {prev_dmg_band} to "
                        f"{cur_dmg_band} ({damage_bands[cur_dmg_band - 1]}"
                        f" < D ≤ {damage_bands[cur_dmg_band]})\n"
                        f"                • current damage value: "
                        f"{cumsum_nl_dmg[i - 1]}\033[0m")

        # Damage weight for the current cycle
        w_ij: float = (
            (damage_bands[cur_dmg_band] - damage_bands[cur_dmg_band - 1])
            / ((damage_bands[cur_dmg_band] ** (1 / dmg_exp[i]))
                - (damage_bands[cur_dmg_band - 1] ** (1 / dmg_exp[i])))
        )
        # Update the damage value for the current cycle
        nl_damage[i] = w_ij * pm_damage[i] if i >= 1 else pm_damage[0]
        cumsum_nl_dmg[i] = (
            nl_damage[i] + cumsum_nl_dmg[i - 1] if i >= 1 else nl_damage[0]
        )
        if i >= 1 and cumsum_nl_dmg[i] > limit_damage:
            w_msg = f"Damage value exceeds {limit_damage} at index {i}"
            warnings.warn(w_msg, UserWarning)
            break
        # fmt: on

    # Edit nl_damage and cumsum_nl_dmg so that all the valus of nldamage[i+1:]
    # are 0 and all the value of cumsum_nl_dmg[i+1:] = cumsum_nl_dmg[i]
    nl_damage[i + 1 :] = 0
    cumsum_nl_dmg[i + 1 :] = cumsum_nl_dmg[i]
    if not plot:
        return nl_damage, cumsum_nl_dmg, None, None
    fig, ax = make_axes()
    cumsum_pm_dmg = np.cumsum(pm_damage)
    fig, ax = _plot_damage_accumulation(
        cumsum_nl_dmg, cumsum_pm_dmg, limit_damage, fig, ax
    )
    return nl_damage, cumsum_nl_dmg, fig, ax


def get_nonlinear_damage(
    damage_rule: str,
    cycle_count: CycleCount,
    sn_curve: SNCurve,
    **kwargs,
) -> float:
    """Calculate the fatigue damage using a nonlinear damage rule.

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Manson-Halford', 'Si-Jian', 'Leve'.
    cycle_count : CycleCount
        The cycle count object.
    sn_curve : SNCurve
        The SN curve object.
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    float
        The cumulated damage.
    """
    if cycle_count.unit != sn_curve.unit:
        e_msg = "The units of the cycle count and SN curve must be the same."
        raise ValueError(e_msg)
    return calc_nonlinear_damage(
        damage_rule,
        cycle_count.stress_range,
        cycle_count.count_cycle,
        sn_curve,
        **kwargs,
    )


def get_nonlinear_damage_with_dca(
    damage_rule: str,
    cycle_count: CycleCount,
    sn_curve: SNCurve,
    damage_bands: np.ndarray,
    limit_damage: Union[float, int] = 1,
    logger: logging.Logger | None = None,
    plot: bool = False,
    **kwargs,
) -> tuple[
    np.ndarray,
    np.ndarray,
    matplotlib.figure.Figure | None,
    matplotlib.axes.Axes | None,
]:
    """Calculate the fatigue damage using a nonlinear damage rule based on the
    Damage Curve Approach (DCA) among the allowed ones:

    - 'Pavlou': Pavlou damage rule
    - 'Theil': Theil damage rule

    Refer to :func:`calc_nonlinear_damage_with_dca` for the
    complete documentation with mathematical details.

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Theil'.
    cycle_count : CycleCount
        The cycle count object.
    sn_curve : SNCurve
        The SN curve object.
    damage_bands : np.ndarray
        The damage bands.
    limit_damage : Union[float, int], optional
        The limit damage, by default 1
    logger : logging.Logger, optional
        The logger object, by default None
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        The cycle-to-cycle damage and the cumulated damage.
    """
    if cycle_count.unit != sn_curve.unit:
        e_msg = "The units of the cycle count and SN curve must be the same."
        raise ValueError(e_msg)
    kwargs["use_dca"] = True
    return calc_nonlinear_damage_with_dca(
        damage_rule,
        cycle_count.stress_range,
        cycle_count.count_cycle,
        sn_curve,
        damage_bands,
        limit_damage,
        logger,
        plot=plot,
        **kwargs,
    )


def find_sn_curve_intersection(
    slope: np.ndarray,
    intercept: np.ndarray,
    endurance: float,
    weight: float,
    res_stress: float,
    n_min: float,
    n_max: float,
    tol=1e-6,
):
    """
    Solve the following nonlinear equation:

    .. math::

        \\left(\\frac{a}{N}\\right)^{\\frac{1}{m}} = w \\cdot N + b

    where:

    - a is the intercept of the S-N curve,
    - m is the slope of the S-N curve,
    - w is the weight for the stress range,
    - b is the residual stress range,
    - N is the number of cycles to failure,

    for N in the range [:math:`n_{min}`, :math:`n_{max}`].
    The equation is linked to the S-N curve definition, therefore it can be
    as complex as the S-N curve itself, i.e. log-linear, log-bilinear, etc.

    Parameters
    ----------
    intercept : float
        The intercept of the S-N curve.
    slope : float
        The slope of the S-N curve.
    endurance : float
        The endurance limit.
    weight : float
        The weight for the stress range.
    res_stress : float
        The residual stress range.
    n_min : float, optional
        The minimum number of cycles to failure, by default 1e0.
    n_max : float, optional
        The maximum number of cycles to failure, by default 1e10.
    tol : float, optional
        The tolerance for the bisection method, by default 1e-6.

    Returns
    -------
    float
        The number of cycles to failure.

    Raises
    ------
    ValueError
        If the bisection method fails to find a solution.
    """
    return numba_bisect(
        __jit_sn_curve_residuals,
        n_min,
        n_max,
        tol,
        1000,
        slope,
        intercept,
        endurance,
        weight,
        res_stress,
    )
    # from ..utils import python_bisect
    # from ..material.sn_curve import __sn_curve_residuals
    # return python_bisect(
    #     __sn_curve_residuals,
    #     n_min,
    #     n_max,
    #     tol,
    #     1000,
    #     slope, intercept, endurance, weight, res_stress
    # )


@nb.njit(
    fastmath=True,
    cache=True,
)
def _calc_theil_sn_damage(stress_range, count_cycle, cycles_to_failure):
    """
    Calculate the cycles to failure using the Theil's method for variable
    amplitude loading.
    Perform a fatigue life prediction under a sequence of stress range blocks,
    taking into account the interaction between different stress levels. It
    uses Theil's method to determine the effective stress range and accumulated
    damage, and an S-N curve to predict the number of cycles to failure.

    Mathematically, the method can be described as follows. Provided a sequence
    of stress range blocks :math:`\\Delta\\sigma_i` and the corresponding
    number of cycles :math:`n_i`, after initializing all the cumulative
    variables to zero, the effective stress range is calculated as:

    .. math::

        N_i & = \\text{SN Curve}\\left(\\Delta\\sigma_i \\right) \\\\
        w_i & = \\frac{\\Delta\\sigma_i}{N_i} \\\\
        \\Delta\\sigma_{eff,i} & = \\Delta\\sigma_{cumsum,i-1}
                                 - w_i \\cdot n_{cumsum, i-1} \\\\
        n_{cumsum,i} & = n_{cumsum,i-1} + n_i \\\\
        \\Delta\\sigma_{cumsum,i} & = w_i \\cdot n_{cumsum,i}
                                    + \\Delta\\sigma_{eff,i}

    where:

    - :math:`\\Delta\\sigma_{eff}` is the effective stress range,
    - :math:`\\Delta\\sigma_{cumsum}` is the cumulative stress range,
    - :math:`w` is the damage weight for the current cycle,
    - :math:`n_{cumsum}` is the cumulative number of cycles,
    - :math:`n` is the number of cycles for the current block.

    The method is applied to each block in the sequence, and the cumulative
    number of cycles is updated at the end of each block. The process continues
    until the number of cycles to failure is determined. If failure does not
    occur until the end of the sequence, the number of cycles to failure is
    estimated by "extrapolating" the last block, i.e., by calculating the
    intersection of the S-N curve with the effective stress range of the last
    block through the bisection method.

    Parameters
    ----------
    stress_range : np.ndarray
        A numpy array containing the stress range for each block (MPa).
    count_cycle : np.ndarray
        A numpy array containing the number of cycles for each block.
    sn_curve : SNCurve
        An object representing the S-N curve of the material.
        It must have attributes `intercept` (float) and `slope` (float).
    plot : bool, optional
        If True, a plot of the damage accumulation is generated. Defaults to
        False.
    Returns
    -------
    tuple
        A tuple containing:

        - n_to_failure (float): The number of cycles to failure. If failure
          occurs within a block, the cumulative number of cycles at the end
          of that block is returned.
        - history (list): A list of tuples, where each tuple contains:

          - n_segment (np.ndarray): An array of cycle numbers for the
            segment.
          - stress_range_segment (np.ndarray): An array of corresponding
            stress ranges for the segment.
          - label (str): A label for the segment, indicating the block
            number and stress range.

    Raises
    ------
    ValueError
        If the fatigue life solver fails to find a solution within a block,
        indicating that failure has occurred within that block.
    Notes
    -----
    - The function assumes that the stress range and count cycle arrays have
      the same length.
    - The Theil's method is used to calculate the effective stress range, which
      takes into account the sequence of stress levels.
    - The S-N curve is used to relate the stress range to the number of cycles
      to failure.
    """

    history = []

    # Start at 0 cycles and 0 stress range
    n_cumsum = 0.0
    stress_range_cumsum = 0.0  # effective stress range at the block end

    # Process predetermined blocks (blocks 1 to L-1)
    w_ij = np.divide(stress_range, cycles_to_failure)
    for i in nb.prange(stress_range.size):  # pylint: disable=E1133
        current_stress_range = stress_range_cumsum - w_ij[i] * n_cumsum
        num_points = int(np.ceil(count_cycle[i]))
        n_end = n_cumsum + count_cycle[i]
        n_segment = np.logspace(
            np.log10(n_cumsum + 1), np.log10(n_end), num_points
        )
        stress_range_segment = w_ij[i] * n_segment + current_stress_range
        # label = f"Block {i+1} (r={stress_range[i]} MPa)"
        history.append((n_segment, stress_range_segment))
        n_cumsum = n_end
        stress_range_cumsum = w_ij[i] * n_cumsum + current_stress_range

    return history, stress_range_cumsum, n_cumsum


def calc_theil_sn_damage(
    stress_range, count_cycle, sn_curve: SNCurve, to_failure: bool = False
) -> tuple[tuple[str, Any, Any], ...]:
    """
    Calculate the cycles to failure using the Theil's method for variable
    amplitude loading.
    Perform a fatigue life prediction under a sequence of stress range blocks,
    taking into account the interaction between different stress levels. It
    uses Theil's method to determine the effective stress range and accumulated
    damage, and an S-N curve to predict the number of cycles to failure.

    Mathematically, the method can be described as follows. Provided a sequence
    of stress range blocks :math:`\\Delta\\sigma_i` and the corresponding
    number of cycles :math:`n_i`, after initializing all the cumulative
    variables to zero, the effective stress range is calculated as:

    .. math::

        N_i & = \\text{SN Curve}\\left(\\Delta\\sigma_i \\right) \\\\
        w_i & = \\frac{\\Delta\\sigma_i}{N_i} \\\\
        \\Delta\\sigma_{eff,i} & = \\Delta\\sigma_{cumsum,i-1}
                                 - w_i \\cdot n_{cumsum, i-1} \\\\
        n_{cumsum,i} & = n_{cumsum,i-1} + n_i \\\\
        \\Delta\\sigma_{cumsum,i} & = w_i \\cdot n_{cumsum,i}
                                    + \\Delta\\sigma_{eff,i}

    where:

    - :math:`\\Delta\\sigma_{eff}` is the effective stress range,
    - :math:`\\Delta\\sigma_{cumsum}` is the cumulative stress range,
    - :math:`w` is the damage weight for the current cycle,
    - :math:`n_{cumsum}` is the cumulative number of cycles,
    - :math:`n` is the number of cycles for the current block.

    The method is applied to each block in the sequence, and the cumulative
    number of cycles is updated at the end of each block. The process continues
    until the number of cycles to failure is determined. If failure does not
    occur until the end of the sequence, the number of cycles to failure is
    estimated by "extrapolating" the last block, i.e., by calculating the
    intersection of the S-N curve with the effective stress range of the last
    block through the bisection method.

    Parameters
    ----------
    stress_range : np.ndarray
        A numpy array containing the stress range for each block (MPa).
    count_cycle : np.ndarray
        A numpy array containing the number of cycles for each block.
    sn_curve : SNCurve
        An object representing the S-N curve of the material.
        It must have attributes `intercept` (float) and `slope` (float).
    Returns
    -------
    tuple
        A tuple containing:

        - n_to_failure (float): The number of cycles to failure. If failure
          occurs within a block, the cumulative number of cycles at the end
          of that block is returned.
        - history (list): A list of tuples, where each tuple contains:

          - n_segment (np.ndarray): An array of cycle numbers for the
            segment.
          - stress_range_segment (np.ndarray): An array of corresponding
            stress ranges for the segment.
          - label (str): A label for the segment, indicating the block
            number and stress range.

    Raises
    ------
    ValueError
        If the fatigue life solver fails to find a solution within a block,
        indicating that failure has occurred within that block.
    Notes
    -----
    - The function assumes that the stress range and count cycle arrays have
      the same length.
    - The Theil's method is used to calculate the effective stress range, which
      takes into account the sequence of stress levels.
    - The S-N curve is used to relate the stress range to the number of cycles
      to failure.
    """
    history, stress_range_cumsum, n_cumsum = _calc_theil_sn_damage(
        count_cycle=count_cycle,
        stress_range=stress_range,
        cycles_to_failure=sn_curve.get_cycles(stress_range),
    )

    labels = [
        f"Block {i + 1} (r={stress_range[i]} MPa)"
        for i in range(len(stress_range))
    ]
    # tuple([(label, *tup) for tup, label in zip(history, labels)])
    hist_dict = {
        label: (n_segment, stress_range_segment)
        for (n_segment, stress_range_segment), label in zip(history, labels)
    }
    if to_failure:
        w_final = stress_range[-1] / sn_curve.get_cycles(stress_range[-1])
        # print("w_final", w_final)
        current_stress_range = stress_range_cumsum - w_final * n_cumsum
        # print("current_stress_range", current_stress_range)
        # Solve modified Basquin for total cycles at final block:
        n_final = find_sn_curve_intersection(
            sn_curve.slope,
            sn_curve.intercept,
            sn_curve.endurance,
            current_stress_range,
            w_final,
            n_min=n_cumsum,
            n_max=1e15,
        )
        n_final_array = np.logspace(np.log10(n_cumsum), np.log10(n_final), 20)
        stress_range_final = w_final * n_final_array + current_stress_range
        # n_cumsum += n_final
        # stress_range_cumsum = w_final * n_cumsum + current_stress_range
        # Append to the last label block
        hist_dict[labels[-1]] = (
            np.append(history[-1][0], n_final_array),
            np.append(history[-1][1], stress_range_final),
        )
    # fmt: off
    return tuple((label, *tup) for tup, label in zip(hist_dict.values(),
                                                      hist_dict.keys()))
    # fmt: on



from .generic import InfiniteSurface, AbstractCrackGeometry
from .cylinder import HollowCylinder, Cylinder, f_hol_cyl_01

# from py_fatigue.geometry import generic, cylinder
# __all__ = ["generic", "cylinder"]

__all__ = [
    "AbstractCrackGeometry",
    "InfiniteSurface",
    "HollowCylinder",
    "Cylinder",
    "f_hol_cyl_01",
]



# -*- coding: utf-8 -*-

r"""The :mod:`py_fatigue.py_fatigue` module collects all functions related to
:term:`cycle-counting(s)<Cycle-counting>` definition.
The main class is CycleCount.
"""

# Standard imports
from __future__ import annotations
import copy
import datetime as dt
import itertools
from dataclasses import dataclass, field
from time import sleep
from types import SimpleNamespace
from typing import (
    Any,
    Sequence,
    cast,
    Callable,
    Iterable,
    Optional,
    Union,
    NewType,
    Tuple,
    TypeVar,
)

# import io
# import itertools
import warnings

# Non-standard imports
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from .rainflow import rainflow as calc_rainflow
from . import histogram as ht
from .. import utils as pfu
from ..mean_stress import corrections as msc
from ..mean_stress import MeanStress
from ..stress_range import StressRange

from ..styling import TermColors, py_fatigue_formatwarning

__all__ = ["CycleCount"]

warnings.formatwarning = py_fatigue_formatwarning

NDT = NewType("NDT", dt.datetime)  # non-aware datetime
ADT = NewType("ADT", dt.datetime)  # timezone aware datetime
DatetimeLike = TypeVar("DatetimeLike", ADT, NDT)


def _assess_json_keys(data: dict) -> dict:
    """Assess the keys in a JSON file.

    Parameters
    ----------
    data : dict
        Dictionary to assess

    Returns
    -------
    dict
        Dictionary with the same keys as the input
    """

    allowed = (
        "bin_lb",
        "range_bin_lower_bound",
        "bin_width",
        "range_bin_width",
        "mean_bin_lower_bound",
        "mean_bin_width",
        "res",
        "no_sm_c",
        "nr_small_cycles",
        "res_sig",
        "hist",
        "lg_c",
        "as_dur",
    )

    for k in data:
        if k not in allowed:
            raise KeyError(f"{k} is not an allowed key")
    if "range_bin_lower_bound" in data:
        if "bin_lb" in data:
            key_err = "'bin_lb' and 'range_bin_lower_bound' are both present"
            raise KeyError(key_err)
    if "bin_lb" in data:
        data["range_bin_lower_bound"] = data.pop("bin_lb")
        data["range_bin_width"] = data.pop("bin_width")

    if "mean_bin_lower_bound" not in data:
        data["mean_bin_lower_bound"] = None

    if "no_sm_c" in data and "nr_small_cycles" not in data:
        data["nr_small_cycles"] = data.pop("no_sm_c")

    # try:
    #     data["res"]
    # except KeyError as key_error:
    #     raise KeyError("'res' is not defined in data") from key_error
    # try:
    #     data["lg_c"]
    # except KeyError as key_error:
    #     raise KeyError("'lg_c' is not defined in data") from key_error
    try:
        data["res_sig"]
    except KeyError:
        data["res_sig"] = None

    return data


def _build_input_data_from_json(  # noqa: C901
    data: dict,
    timestamp: Union[ADT, NDT] = ADT(dt.datetime.now(dt.timezone.utc)),
    round_decimals: int = 4,
    name: Optional[str] = None,
    mean_stress_corrected: str = "No",
    lffd_solved: bool = False,
) -> dict:
    """Function returning the same data it receives"""
    data = _assess_json_keys(data)
    the_hist = np.empty(0)
    range_bin_centers = np.empty(0)
    mean_bin_centers = np.empty(0)
    if "hist" not in data:
        data["hist"] = []
    if (
        len(data["hist"]) == 0
        and len(data["res"]) == 0
        and len(data["lg_c"]) == 0
    ):
        raise ValueError("'hist' is empty")
    for k, v in data.items():
        if k != "hist":
            data[k] = np.array(v) if isinstance(v, list) else v
    for hst in data["hist"]:
        hst = np.array(hst)
    # A) Case without mean stress
    # Adding hist cycles
    if len(data["hist"]) > 0 and not isinstance(data["res"][0], Iterable):
        the_hist = np.asarray(data["hist"])
        zero_hist_counts = np.where(the_hist == 0)[0]
        range_bin_centers = np.around(
            np.linspace(
                data["range_bin_lower_bound"] + data["range_bin_width"] / 2,
                data["range_bin_lower_bound"]
                + data["range_bin_width"] / 2
                + data["range_bin_width"] * (len(data["hist"]) - 1),
                len(data["hist"]),
            ),
            round_decimals,
        )
        the_hist = np.delete(the_hist, zero_hist_counts)
        range_bin_centers = np.delete(range_bin_centers, zero_hist_counts)
    # Adding large cycles
    if len(data["lg_c"]) > 0 and not isinstance(
        data["res"][0],
        Iterable,  # isinstance(
        # data["lg_c"][0], (float, int, np.int16, np.int32)
    ):
        the_hist = np.hstack([the_hist, np.ones(len(data["lg_c"]))])
        range_bin_centers = np.hstack([range_bin_centers, data["lg_c"]])
    # Adding residual cycles
    if len(data["res"]) > 0 and not isinstance(data["res"][0], Iterable):
        the_hist = np.hstack([the_hist, 0.5 * np.ones(len(data["res"]))])
        range_bin_centers = np.hstack([range_bin_centers, data["res"]])
    # B) Case with mean stress
    if len(data["hist"]) > 0 and isinstance(data["hist"][0], Iterable):
        the_hist = np.asarray(list(itertools.chain.from_iterable(data["hist"])))
        zero_hist_counts = np.where(the_hist == 0)[0]
        for idx in range(len(data["hist"])):
            len_mean_bin = len(data["hist"][idx])
            new_ranges = np.around(
                np.linspace(
                    data["range_bin_lower_bound"] + data["range_bin_width"] / 2,
                    data["range_bin_lower_bound"]
                    + data["range_bin_width"] / 2
                    + data["range_bin_width"] * (len_mean_bin - 1),
                    len_mean_bin,
                ),
                round_decimals,
            )
            range_bin_centers = np.hstack([range_bin_centers, new_ranges])
            new_means = (
                data["mean_bin_lower_bound"]
                + idx * data["mean_bin_width"] * np.ones(len_mean_bin)
                + data["mean_bin_width"] / 2
            )
            mean_bin_centers = np.hstack([mean_bin_centers, new_means])

        the_hist = np.delete(the_hist, zero_hist_counts)
        range_bin_centers = np.delete(range_bin_centers, zero_hist_counts)
        mean_bin_centers = np.delete(mean_bin_centers, zero_hist_counts)
    # Adding large cycles
    if len(data["lg_c"]) > 0 and isinstance(data["lg_c"][0], Iterable):
        the_hist = np.hstack([the_hist, np.ones(len(data["lg_c"][:, 0]))])
        mean_bin_centers = np.hstack([mean_bin_centers, data["lg_c"][:, 0]])
        range_bin_centers = np.hstack([range_bin_centers, data["lg_c"][:, 1]])
    # Adding residual cycles
    if len(data["res"]) > 0 and isinstance(data["res"][0], Iterable):
        the_hist = np.hstack([the_hist, 0.5 * np.ones(len(data["res"][:, 0]))])
        mean_bin_centers = np.hstack([mean_bin_centers, data["res"][:, 0]])
        range_bin_centers = np.hstack([range_bin_centers, data["res"][:, 1]])

    df = {
        "count_cycle": the_hist,
        "mean_stress": (
            mean_bin_centers
            if "mean_bin_width" in data
            else np.zeros(len(range_bin_centers))
        ),
        "stress_range": range_bin_centers,
        "range_bin_lower_bound": data["range_bin_lower_bound"],
        "range_bin_width": data["range_bin_width"],
        "_mean_bin_lower_bound": (
            data["mean_bin_lower_bound"]
            if "mean_bin_lower_bound" in data
            else None
        ),
        "mean_bin_width": (
            data["mean_bin_width"] if "mean_bin_width" in data else 10
        ),
        "nr_small_cycles": data["nr_small_cycles"],
        "residuals_sequence": data["res_sig"],
        "timestamp": timestamp,
        "name": name,
        "mean_stress_corrected": mean_stress_corrected,
        "lffd_solved": lffd_solved,
    }
    return df


@dataclass(repr=False)
class CycleCount:
    """Cycle-counting class.

    The class heavily relies on the :mod:`py_fatigue.rainflow` and
    :mod:`py_fatigue.histogram` modules.

    CycleCount instances should be defined from either of the
    following two class methods:

        - :func:`~CycleCount.from_rainflow`
        - :func:`~CycleCount.from_timeseries`
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    # 16 is reasonable in this case.

    count_cycle: np.ndarray[Any, np.dtype[np.float64]]
    stress_range: np.ndarray[Any, np.dtype[np.float32]]
    mean_stress: np.ndarray[Any, np.dtype[np.float64]]
    timestamp: Union[ADT, NDT] = ADT(dt.datetime.now(dt.timezone.utc))
    unit: str = "MPa"
    range_bin_lower_bound: float = 0.2
    range_bin_width: float = 0.05
    _mean_bin_lower_bound: Optional[float] = None
    mean_bin_width: float = 10
    nr_small_cycles: float = 0
    residuals_sequence: np.ndarray[Any, Any] = field(
        default_factory=lambda: np.empty(0)
    )
    _min_max_sequence: np.ndarray[Any, Any] = field(
        default_factory=lambda: np.empty(0)
    )
    lffd_solved: bool = False
    mean_stress_corrected: str = "No"
    stress_concentration_factor: float = 1.0
    name: Optional[str] = None
    _time_sequence: Optional[np.ndarray] = None

    def __post_init__(self):
        if not np.isnan(np.min(self.mean_stress)):
            if self._mean_bin_lower_bound is None:
                self._mean_bin_lower_bound = (
                    round(min(self.mean_stress) / self.mean_bin_width)
                    * self.mean_bin_width
                    - self.mean_bin_width / 2
                )
            if self.time_sequence is None:
                self.time_sequence = np.asarray(self.timestamp)

    @classmethod
    def from_rainflow(
        cls,
        data: dict,
        timestamp: Union[ADT, NDT] = ADT(dt.datetime.now(dt.timezone.utc)),
        round_decimals: int = 4,
        name: Optional[str] = None,
        mean_stress_corrected: str = "No",
        lffd_solved: bool = False,
        unit: str = "MPa",
    ) -> "CycleCount":
        """Create a cycle-counting object from rainflow cycles.

        Parameters
        ----------
        *args
            Positional arguments passed to `py_fatigue.rainflow.rainflow`.
        **kwargs
            Keyword arguments passed to `py_fatigue.rainflow.rainflow`.

        Returns
        -------
        CycleCount
            A cycle-counting object.
        """
        df = _build_input_data_from_json(
            data,
            timestamp,
            round_decimals,
            name,
            mean_stress_corrected,
            lffd_solved,
        )
        df["unit"] = unit
        return cls(**df)

    @classmethod
    def from_timeseries(
        cls,
        data: np.ndarray,
        time: Optional[np.ndarray] = None,
        timestamp: Union[ADT, NDT] = ADT(dt.datetime.now(dt.timezone.utc)),
        range_bin_lower_bound: float = 0.2,
        range_bin_width: float = 0.05,
        mean_bin_lower_bound: Optional[float] = None,
        mean_bin_width: float = 10,
        name: Optional[str] = None,
        unit: str = "MPa",
    ) -> "CycleCount":
        """Generate a cycle-count from a timeseries.

        Parameters
        ----------
        *args
            Positional arguments passed to `_build_input_data_from_json`.
        **kwargs
            Keyword arguments passed to `_build_input_data_from_json`.
        """
        cc = calc_rainflow(data=data, time=time, extended_output=True)
        stress_range = 2 * cc[0][:, 0]
        mean_stress = cc[0][:, 1]
        count_cycle = cc[0][:, 2]
        residuals_sequence = cc[1]
        df = {
            "count_cycle": count_cycle,
            "mean_stress": mean_stress,
            "stress_range": stress_range,
            "timestamp": timestamp,
            "range_bin_lower_bound": range_bin_lower_bound,
            "range_bin_width": range_bin_width,
            "_mean_bin_lower_bound": mean_bin_lower_bound,
            "mean_bin_width": mean_bin_width,
            "residuals_sequence": residuals_sequence,
            "name": name,
            "unit": unit,
        }
        return cls(**df)

    # ! Instance managed attributes
    @property
    def mean_bin_lower_bound(self) -> float:
        """The stress values.

        Returns
        -------
        values : np.ndarray
        """
        return cast(float, self._mean_bin_lower_bound)

    @mean_bin_lower_bound.setter
    def mean_bin_lower_bound(self, value: float) -> None:
        self._mean_bin_lower_bound = value

    @property
    def time_sequence(self) -> np.ndarray:
        """The sequence of timestamps possibly added.

        Returns
        -------
        time_sequence : np.ndarray
        """
        if self._time_sequence is None:
            return np.asarray([self.timestamp])
        return self._time_sequence

    @property
    def min_max_sequence(self) -> np.ndarray:
        """The minimum and maximum of the :term:`residuals<Residuals>`
        (half-cycles) sequence.

        Returns
        -------
        np.ndarray, shape=(2, )
            The minimum and maximum of the half-cycles sequence.
        """
        if self.residuals_sequence is None:
            return self._min_max_sequence
        if (
            len(self._min_max_sequence) == 0
            and len(self.residuals_sequence) >= 2
        ):
            return np.hstack(
                [
                    min(self.residuals_sequence),
                    max(self.residuals_sequence),
                ]
            )
        return self._min_max_sequence

    @property
    def residuals(self) -> np.ndarray:
        """Extracts the :term:`residuals<Residuals>`, also known as half
        cycles, from mean-range couples of the cycle-counting object.

        Returns
        -------
        np.ndarray, shape (n, 2)
            Array containing mean stresses and stress ranges of the residuals.
            - residual_mean_stresses = array[:, 0]
            - residual_stress_ranges = array[:, 1]
        """
        _, idx = pfu.split_full_cycles_and_residuals(self.count_cycle)
        return np.asarray([self.mean_stress[idx], self.stress_range[idx]]).T

    @property
    def full_cycles(self) -> np.ndarray:
        """Extracts the full cycles only from mean-range couples.

        Returns
        -------
        np.ndarray, shape (n, 2)
            Array containing mean stresses and stress ranges of the full
            cycles.
            - full_cycles_mean_stresses = array[:, 0]
            - full_cycles_stress_ranges = array[:, 1]
        """
        full_mean = pfu.calc_full_cycles(self.mean_stress, self.count_cycle)
        full_range = pfu.calc_full_cycles(self.stress_range, self.count_cycle)
        return np.asarray([full_mean, full_range]).T

    @property
    def half_cycles(self) -> np.ndarray:
        """Symlink to residuals

        See also
        --------
        :func:`py_fatigue.cycle_count.CycleCount.residuals`
        """
        return self.residuals

    @property
    def mean_bin_upper_bound(self) -> float:
        """Returns the upper bound of the mean stress bin edges.

        Returns
        -------
        float
            The upper bound of the mean stress bins (highest mean bin edge).
        """
        return pfu.bin_upper_bound(
            self.mean_bin_lower_bound,
            self.mean_bin_width,
            max(self.full_cycles[:, 0]),
        )

    @property
    def range_bin_upper_bound(self) -> float:
        """Returns the upper bound of the stress range bins.

        Returns
        -------
        float
            The upper bound of the stress range bins (highest range bin edge).
        """
        return pfu.bin_upper_bound(
            self.range_bin_lower_bound,
            self.range_bin_width,
            max(self.full_cycles[:, 1]),
        )

    @property
    def bin_centers(self) -> tuple:
        """Extracts the bin centers from the cycle-counting object.

        Returns
        -------
        tuple
            Tuple containing mean stresses and stress ranges of the bin
            centers.
            - bin_centers_mean_stresses = tuple[0]
            - bin_centers_stress_ranges = tuple[1]
        """
        mean_bin_centers = (
            pfu.calc_bin_edges(
                self.mean_bin_lower_bound,
                self.mean_bin_upper_bound,
                self.mean_bin_width,
            )[:-1]
            + self.mean_bin_width / 2
        )

        range_bin_centers = (
            pfu.calc_bin_edges(
                self.range_bin_lower_bound,
                self.range_bin_upper_bound,
                self.range_bin_width,
            )[:-1]
            + self.range_bin_width / 2
        )
        return mean_bin_centers, range_bin_centers

    @property
    def bin_edges(self) -> tuple:
        """Extracts the bin edges from the cycle-counting object.

        Returns
        -------
        ht.tuple
            Tuple containing mean stresses and stress ranges of the bin edges.
            - bin_edges_mean_stresses = tuple[0]
            - bin_edges_stress_ranges = tuple[1]
        """
        mean_bin_edges = pfu.calc_bin_edges(
            self.mean_bin_lower_bound,
            self.mean_bin_upper_bound,
            self.mean_bin_width,
        )

        range_bin_edges = pfu.calc_bin_edges(
            self.range_bin_lower_bound,
            self.range_bin_upper_bound,
            self.range_bin_width,
        )
        return mean_bin_edges, range_bin_edges

    @property
    def stress_amplitude(self) -> np.ndarray:
        """Extracts the stress amplitude from the cycle-counting object.

        Returns
        -------
        np.ndarray
            Array containing the stress amplitude.
        """
        return self.stress_range / 2

    @property
    def min_stress(self) -> np.ndarray:
        """Extracts the min stress from the cycle-counting object.

        Returns
        -------
        np.ndarray
            Array containing the min stresses.
        """
        return self.mean_stress - self.stress_amplitude

    @property
    def max_stress(self) -> np.ndarray:
        """Extracts the max stress from the cycle-counting object.

        Returns
        -------
        np.ndarray
            Array containing the max stresses.
        """
        return self.mean_stress + self.stress_amplitude

    @property
    def statistical_moments(self) -> tuple[float, float, float]:
        """Calculate the spectral moments from the cycle-counting object.

        Returns
        -------
        tuple[float, float, float]
            Tuple containing the spectral moments.
            - mean = tuple[0]
            - coefficient of variance = tuple[1]
            - skewness = tuple[2]
        """
        stress_range = np.hstack(
            [self.half_cycles[:, 1], self.full_cycles[:, 1]]
        )
        m_1 = np.mean(stress_range)
        std_dev = np.sqrt(np.mean((stress_range - np.mean(stress_range)) ** 2))
        m_2 = std_dev / m_1
        m_3 = np.mean((stress_range - np.mean(stress_range)) ** 3) / (m_1**3)

        return m_1, m_2, m_3

    # ! Representations
    def as_dict(
        self,
        damage_tolerance_for_binning: float = 0.001,
        damage_exponent: float = 5.0,
        round_decimals: int = 4,
        max_consecutive_zeros: int = 32,
        legacy_export: bool = False,
        debug_mode=False,
    ) -> dict:
        """Convert the cycle-counting object to a dictionary.

        Parameters
        ----------
        damage_tolerance_for_binning : float, optional
            Tolerance for large-cycles calculation when binning,
            by default 0.001
        damage_exponent : float, optional
            SN curve exponent used to define large cycles. Higher means more
            conservative, by default 5.0
        max_consecutive_zeros : int, optional
            maximum number of consecutive zeros before saving a cycle
            to large cycles array. This quantity is introduced to save
            storage space, by default 8
        round_decimals : int, optional
            Number of decimals after comma, by default 4

        Returns
        -------
        dict
            Dictionary containing the following keys:
            - "hist"
            - "range_bin_lower_bound"
            - "range_bin_width"
            - "mean_bin_lower_bound"
            - "mean_bin_width"
            - "nr_small_cycles"
            - "res_sig"
            - "res"
            - "lg_c"

        See also
        --------
        :func:`py_fatigue.cycle_count.rainflow_binner`
        """
        stress_ranges = StressRange(
            _counts=self.count_cycle,
            _values=self.stress_range,
            bin_width=self.range_bin_width,
            _bin_lb=self.range_bin_lower_bound,
        )

        mean_stresses = MeanStress(
            _counts=self.count_cycle,
            _values=self.mean_stress,
            bin_width=self.mean_bin_width,
            _bin_lb=self.mean_bin_lower_bound,
        )

        output_dict = ht.rainflow_binner(
            stress_ranges,
            mean_stresses if not legacy_export else None,
            damage_tolerance_for_binning=damage_tolerance_for_binning,
            damage_exponent=damage_exponent,
            max_consecutive_zeros=max_consecutive_zeros,
            round_decimals=round_decimals,
            debug_mode=debug_mode,
        )
        output_dict["res_sig"] = np.asarray(
            np.round(self.residuals_sequence, round_decimals)
        ).tolist()
        return output_dict

    def _repr_html_(self) -> str:  # pragma: no cover
        """HTML representation of the cycle-counting object.

        Returns
        -------
        str
        """
        df = pd.DataFrame()
        df["Cycle counting object"] = [
            f"largest full stress range, {self.unit}",
            f"largest stress range, {self.unit}",
            "number of full cycles",
            "number of residuals",
            "number of small cycles",
            "stress concentration factor",
            "residuals resolved",
            "mean stress-corrected",
        ]

        df[self.name] = [
            (
                max(self.full_cycles[:, 1])
                if max(self.count_cycle) > 0.5
                and max(self.full_cycles[:, 1]) > self.range_bin_lower_bound
                else None
            ),
            max(self.stress_range),
            int(sum(self.count_cycle) - int(len(self.residuals[:, 1])) / 2),
            int(len(self.residuals[:, 1])),
            int(self.nr_small_cycles),
            (
                "N/A"
                if self.stress_concentration_factor == 1
                else self.stress_concentration_factor
            ),
            bool(self.lffd_solved),
            self.mean_stress_corrected,
        ]
        df = df.set_index("Cycle counting object")

        return df._repr_html_()  # pylint: disable=protected-access

    def __str__(self) -> str:
        """String representation of the cycle-counting object.

        Returns
        -------
        str
        """
        strng = f"CC_{self.name}"
        if len(self.time_sequence) == 1:
            strng = "".join(
                (
                    f"CC_{self.name} (",
                    f"{self.timestamp.strftime('%d %b %Y, %H:%M')} - ",
                    "",
                )
            )
        if len(self.time_sequence) > 1:
            strng = "".join(
                (
                    f"CC_ {self.name} (from ",
                    f"{self.time_sequence[0].strftime('%d %b %Y, %H:%M')} to ",
                    f"{self.time_sequence[-1].strftime('%d %b %Y, %H:%M')})",
                )
            )
        return str(strng)

    def to_df(self) -> pd.DataFrame:  # pragma: no cover
        """
        Represent the :class:`CycleCount` instance as a pandas
        dataframe.

        Returns
        -------
        :class:`pd.DataFrame`
            dataframe containing relevant fatigue information
        """

        df = pd.DataFrame.from_dict(
            {
                "count_cycle": self.count_cycle,
                "mean_stress": self.mean_stress,
                "stress_range": self.stress_range,
            }
        )
        df._metadata = {
            "name": self.name,
            "timestamp": self.timestamp,
            "time_sequence": self.time_sequence,
            "residuals_sequence": self.residuals_sequence,
            "mean_stress_corrected": self.mean_stress_corrected,
            "stress_concentration_factor": self.stress_concentration_factor,
            "nr_small_cycles": self.nr_small_cycles,
            "lffd_solved": self.lffd_solved,
            "unit": self.unit,
        }

        return df

    # ! Operators
    def __eq__(self, other: object) -> bool:
        """Equality operation for :class:`CycleCount` instances.

        Parameters
        ----------
        other : object
            object to compare to

        Returns
        -------
        bool
            True if the objects are equal, False otherwise
        """
        if not isinstance(other, CycleCount):
            return False

        return pfu.compare(self.__dict__, other.__dict__)

    def __add__(  # fgh - noqa: _C901_
        self, other: "CycleCount"
    ) -> "CycleCount":
        """Sum operation for :class:`CycleCount` instances.

        Parameters
        ----------
        other : Union[CycleCount, None]
            CycleCount object to be added to the current object.

        Returns
        -------
        CycleCount
            Summed cycle-count object.

        Raises
        ------
        TypeError
            Trying to add a non-CycleCount object.
        TypeError
            Trying to add a CycleCount object with different names.
        ValueError
            Timestamps must be defined for CycleCount sum.
        UserWarning
            Different bin widths.
        TypeError
            Timestamps must be sorted ascending.
        UserWarning
            Adding-up instances having different stress concentration factors.
        TypeError
            residuals_sequence is not iterable.
        UserWarning
            No residuals_sequence attribute found.
        """

        # Running initial checks
        if not _cycle_count_add_checks(self, other):
            return self

        # Performing actual sum
        # added_counts = np.hstack([self.count_cycle, other.count_cycle])
        # added_range = np.hstack([self.stress_range, other.stress_range])
        # added_mean = np.hstack([self.mean_stress, other.mean_stress])
        (
            added_mean_bin_lower_bound,
            added_mean_bin_width,
            added_range_bin_lower_bound,
            added_range_bin_width,
        ) = _handling_different_bins_in_sum(self, other)

        # added_nr_small_cycles = self.nr_small_cycles + other.nr_small_cycles
        # added_timestamp = np.hstack([self.timestamp, other.timestamp])
        added_scf = np.average(
            [
                self.stress_concentration_factor,
                other.stress_concentration_factor,
            ]
        )
        # added_as_duration = self.as_duration + other.as_duration
        # ? Residuals sequence management
        # 1) Residuals sequence is defined in both self and other
        if hasattr(self, "residuals_sequence") and hasattr(
            other, "residuals_sequence"
        ):
            # 1a) Residuals sequence is ok in both self and other
            if isinstance(self.residuals_sequence, Iterable) and isinstance(
                other.residuals_sequence, Iterable
            ):
                added_res_seq = np.hstack(
                    [self.residuals_sequence, other.residuals_sequence]
                )
                added_min_max = np.hstack(
                    [self.min_max_sequence, other.min_max_sequence]
                )
            # 1b) Residuals sequence is ok in self, but not in other
            elif isinstance(
                self.residuals_sequence, Iterable
            ) and not isinstance(other.residuals_sequence, Iterable):
                w_msg = "".join(
                    (
                        f"No residuals_sequence found on {TermColors.CRED}",
                        f"{other.timestamp.strftime('%d %B %Y, %H:%M')}",
                        f"{TermColors.CYELLOW2}.",
                    )
                )
                warnings.warn(w_msg, UserWarning)
                added_res_seq = self.residuals_sequence
                added_min_max = self.min_max_sequence
            # 1c) Residuals sequence is ok in other, but not in self
            elif not isinstance(
                self.residuals_sequence, Iterable
            ) and isinstance(other.residuals_sequence, Iterable):
                w_msg = "".join(
                    (
                        f"No residuals_sequence found on {TermColors.CRED}",
                        f"{self.timestamp.strftime('%d %B %Y, %H:%M')}",
                        f"{TermColors.CYELLOW2}.",
                    )
                )
                warnings.warn(w_msg, UserWarning)
                added_res_seq = other.residuals_sequence
                added_min_max = other.min_max_sequence
            # 1d) Residuals sequence is neither defined in self nor in other
            else:
                w_msg = "No residuals_sequence attribute found."
                warnings.warn(w_msg, UserWarning)
                added_res_seq = np.empty(0)
                added_min_max = np.empty(0)
        # 2) Residuals sequence is defined in self but not in other
        elif hasattr(self, "residuals_sequence") and not hasattr(
            other, "residuals_sequence"
        ):
            w_msg = "".join(
                (
                    f"No residuals_sequence found on {TermColors.CRED}",
                    f"{other.timestamp.strftime('%d %B %Y, %H:%M')}",
                    f"{TermColors.CYELLOW2}.",
                )
            )
            warnings.warn(w_msg, UserWarning)
            added_res_seq = self.residuals_sequence
            added_min_max = self.min_max_sequence
        # 3) Residuals sequence is defined in other but not in self
        elif not hasattr(self, "residuals_sequence") and hasattr(
            other, "residuals_sequence"
        ):
            w_msg = "".join(
                (
                    f"No residuals_sequence found on {TermColors.CRED}",
                    f"{self.timestamp.strftime('%d %B %Y, %H:%M')}",
                    f"{TermColors.CYELLOW2}.",
                )
            )
            warnings.warn(w_msg, UserWarning)
            added_res_seq = other.residuals_sequence
            added_min_max = other.min_max_sequence
        # 4) Neither self nor other have Residuals sequence
        else:
            w_msg = "No residuals_sequence attribute found."
            warnings.warn(w_msg, UserWarning)
            added_res_seq = np.empty(0)
            added_min_max = np.empty(0)

        added_cc = CycleCount(
            count_cycle=np.hstack([self.count_cycle, other.count_cycle]),
            mean_stress=np.hstack([self.mean_stress, other.mean_stress]),
            stress_range=np.hstack([self.stress_range, other.stress_range]),
            range_bin_lower_bound=added_range_bin_lower_bound,
            range_bin_width=added_range_bin_width,
            _mean_bin_lower_bound=added_mean_bin_lower_bound,
            mean_bin_width=added_mean_bin_width,
            nr_small_cycles=self.nr_small_cycles + other.nr_small_cycles,
            residuals_sequence=added_res_seq,
            _min_max_sequence=added_min_max,
            lffd_solved=False,
            mean_stress_corrected="No",
            stress_concentration_factor=float(added_scf),
            timestamp=min(self.timestamp, other.timestamp),
            name=self.name,
            _time_sequence=np.unique(
                np.hstack([self.time_sequence, other.time_sequence])
            ),
        )
        return added_cc

    def __radd__(self, other):
        """
        Reverse add function.
        """
        if other is None or other == 0 or np.isnan(other):
            return self
        return self.__add__(other)

    def __mul__(self, other: Union[float, int]):
        """Left multiplication of :class:`CycleCount` \
        instance by a scalar stress multiplication factor .

        Parameters
        ----------
        other : Union[float, int]
            stress concentration factor

        Raises
        ------
        TypeError: Multiplication by a scalar is the only thing \
        allowed

        Returns
        -------
        :class:`CycleCount`
            :class:`CycleCount` instance with \
        "raised" stresses and stress concentration factor != 1
        """
        return _multiplication_by_scalar(self, other)

    def __rmul__(self, other):
        """Left multiplication of :class:`CycleCount` \
        instance by a scalar stress multiplication factor.

        Parameters
        ----------
        other : Union[float, int]
            stress concentration factor

        Raises
        ------
        TypeError: Multiplication by a scalar is the only thing \
        allowed

        Returns
        -------
        :class:`CycleCount`
            :class:`CycleCount` instance with \
        "raised" stresses and stress concentration factor != 1
        """
        return _multiplication_by_scalar(self, other)

    def solve_lffd(self, solve_mode: str = "Residuals") -> "CycleCount":
        """Resolve the LFFD using Marsh et al. (2016) method.

        Parameters
        ----------
        solve_mode : str
            Solve mode. Options are: "Residuals" and "Min-Max".

        Returns
        -------
        :class:`CycleCount`
            :class:`CycleCount` instance with :term:`LFFD` resolved.
        """
        return _solve_lffd(self, solve_mode)

    def resolve_residuals(self) -> "CycleCount":
        """Resolve the residuals. It is a symlink to :meth:`solve_lffd`.

        Returns
        -------
        :class:`CycleCount`
            :class:`CycleCount` instance with :term:`LFFD` resolved.

        See also
        --------
        :meth:`solve_lffd`
        """
        return self.solve_lffd()

    def mean_stress_correction(  # pylint: disable=R0912
        self,
        correction_type: str = "DNVGL-RP-C203",
        plot: bool = False,
        enforce_pulsating_load: bool = False,
        **kwargs,
    ) -> "CycleCount":
        """Calculates the mean stress correction returning a new
        :class:`CycleCount` instance with corrected stress ranges
        at constant load ratio (:math:`R=\\sigma_{min}/\\sigma_{max}`).

        Parameters
        ----------
        correction_type : str, optional
            Type of mean stress correction, by default "DNVGL-RP-C203"
        plot : bool, optional
            Plot the mean stress corrected history, by default False
        enforce_pulsating_load : bool, optional
            Choosing to enforce load ratio equals zero prior
            correction (useful in legacy analyses), by default False

        `correction_types` : list
            * *dnvgl-rp-c203*
            * *Walker*
            * *SWT* or *Smith-Watson-Topper*
            * *Goodman*
            * *Gerber*
            * *Morrow*
            * *Soderberg*
            * *Generic-Haigh*

        **kwargs : dict
            *detail_factor* : float
                See DNVGL-RP-C203, by default 0.8
            *yield_strength* : float
                Yielding stress amplitude
            *gamma* : float
                Exponent for Walker mean stress correction, by default 0.5
            *r_out* : float
                Load ratio at which the correction is performed
            *ult_s* : float
                Ultimate tensile strength for Goodman and Gerber, yield
                strength for Soderberg, and true fracture strength for Morrow
            *correction_exponent* : float
                Exponent for Generic-Haigh mean stress
                correction. It is set to 1 for Goodman, Morrow, and Soderberg,
                and 2 for Gerber.


        Returns
        -------
        CycleCount
            :class:`CycleCount` instance with corrected stress ranges

        Raises
        ------
        ValueError
            `correction_type` must be one of `correction_types`.
        ValueError
            Mean stress correction will not be applied because the
            initial mean stress array is made of zeros. Set
            enforce_load_ratio to True to proceed.
        ValueError
            This correction type is not supported.
        ValueError
            The mean stress correction method requires the 'r_out'
            parameter, i.e. the fixed load ratio at which the correction
            is performed.
        ValueError
            The mean stress correction method requires the 'ult_s'
            parameter, i.e. the ultimate tensile strength for Goodman and
            Gerber, the yield strength for Soderberg, and the true fracture
            strength for Morrow.
        ValueError
            The mean stress correction method requires the
            'correction_exponent' parameter, i.e. the exponent of the
            correction function.

        See also
        --------
        :func:`py_fatigue.mean_stress.corrections.dnvgl_mean_stress_correction`
        :func:`py_fatigue.mean_stress.corrections.walker_mean_stress_correction`
        :func:`py_fatigue.mean_stress.corrections.swt_mean_stress_correction`
        :func:`py_fatigue.mean_stress.corrections.goodman_haigh_mean_stress_correction`
        """
        correction_types = [
            "dnvgl-rp-c203",
            "walker",
            "swt",
            "smith-watson-topper",
            "goodman",
            "gerber",
            "morrow",
            "soderberg",
            "generic-haigh",
            "haigh",
        ]
        if correction_type.lower() not in correction_types:
            raise ValueError(
                f"'correction_type' must be one of {correction_types}"
            )
        if self.mean_stress_corrected != "No":
            w_msg = "".join(
                (
                    "Cyclecount instance has already been corrected using ",
                    f"{TermColors.CBEIGE2}{self.mean_stress_corrected}",
                    TermColors.CYELLOW2,
                    ".",
                )
            )
            warnings.warn(w_msg, UserWarning)
            return self

        self_ = copy.deepcopy(self) if self.lffd_solved else self.solve_lffd()

        if np.all((self_.mean_stress == 0)) and not enforce_pulsating_load:
            e_msg = "".join(
                (
                    f"{correction_type} mean stress correction will not be ",
                    "applied because the mean stress array is made of zeros. ",
                    "Set 'enforce_pulsating_load=True' to override mean ",
                    "stresses to pulsating loading (zero load ratio).",
                )
            )
            raise ValueError(e_msg)
        if np.all((self_.mean_stress == 0)) and enforce_pulsating_load:
            self_.mean_stress = self_.stress_amplitude

        ns = SimpleNamespace(**kwargs)
        if correction_type.lower() == ("dnvgl-rp-c203"):
            if "detail_factor" not in kwargs:
                w_msg = "".join(
                    (
                        "No detail factor provided for DNVGL-RP-C203 ",
                        "mean stress correction. \nPassing detail_factor=0.8 ",
                        "This value is suitable for welded connections ",
                        "without significant residual stresses.",
                    )
                )
                warnings.warn(w_msg, UserWarning)
                ns.detail_factor = 0.8
            stress_ranges_corr = msc.dnvgl_mean_stress_correction(
                self_.mean_stress,
                self_.stress_amplitude,
                ns.detail_factor,
                plot,
            )
            mean_stress_corr = copy.deepcopy(stress_ranges_corr / 2)

        if correction_type.lower() == ("walker"):
            if "gamma" not in kwargs:
                w_msg = "".join(
                    (
                        "No gamma exponent provided for Walker ",
                        "mean stress correction. \nPassing gamma=0.5. ",
                        "This value is suitable for many materials ",
                        "and corresponds to applying SWT correction.",
                    )
                )
                warnings.warn(w_msg, UserWarning)
                ns.gamma = 0.5
            stress_ranges_corr = msc.walker_mean_stress_correction(
                self_.mean_stress,
                self_.stress_amplitude,
                ns.gamma,
                plot,
            )
            mean_stress_corr = np.zeros(len(self_.mean_stress))

        if correction_type.lower() in ["swt", "smith-watson-topper"]:
            stress_ranges_corr = msc.swt_mean_stress_correction(
                self_.mean_stress,
                self_.stress_amplitude,
                plot,
            )
            mean_stress_corr = np.zeros(len(self_.mean_stress))
        if correction_type.lower() in [
            "goodman",
            "gerber",
            "morrow",
            "soderberg",
            "haigh",
            "generic-haigh",
        ]:
            if "r_out" not in kwargs:
                e_msg = "".join(
                    (
                        "The mean stress correction method ",
                        f"{correction_type.upper()} requires the ",
                        "'r_out', i.e. the fixed load ratio at which the ",
                        "correction is performed. ",
                    )
                )
                raise ValueError(e_msg)
            if "ult_s" not in kwargs:
                e_msg = "".join(
                    (
                        "The mean stress correction method ",
                        f"{correction_type.upper()} requires the ",
                        "'ult_s' parameter, i.e. the ultimate tensile "
                        "strength for Goodman and Gerber, the yield strength "
                        "for Soderberg, and the true fracture strength for "
                        "Morrow.",
                    )
                )
                raise ValueError(e_msg)
            if correction_type.lower() in ["goodman", "morrow", "soderberg"]:
                ns.correction_exponent = 1
            if correction_type.lower() in ["gerber"]:
                ns.correction_exponent = 2
            if correction_type.lower() in ["generic-haigh", "haigh"]:
                if "correction_exponent" not in kwargs:
                    e_msg = "".join(
                        (
                            "The mean stress correction method ",
                            f"{correction_type.upper()} requires the ",
                            "'correction_exponent' parameter, i.e. the ",
                            "exponent of the correction function.",
                        )
                    )
                    raise ValueError(e_msg)
            if isinstance(ns.r_out, (list, np.ndarray)) and len(ns.r_out) > 1:
                w_msg = "".join(
                    (
                        "Multiple output load ratios provided!\n",
                        f"Only the first one (r_out={ns.r_out[0]}) will be "
                        "used in the mean stress-corrected CycleCount "
                        "instance creation, while the others will be used "
                        "for plotting purposes.",
                    )
                )
                warnings.warn(w_msg, UserWarning)
                if not plot:
                    ns.r_out = ns.r_out[0]
            s_a_c, mean_stress_corr = msc.goodman_haigh_mean_stress_correction(
                amp_in=self_.stress_amplitude,
                mean_in=self_.mean_stress,
                r_out=ns.r_out,
                ult_s=ns.ult_s,
                correction_exponent=ns.correction_exponent,
                plot=plot,
            )
            stress_ranges_corr = s_a_c[0] * 2
            mean_stress_corr = mean_stress_corr[0]

        # ! Creating the corrected CycleCount instance
        cc_msc = CycleCount(
            count_cycle=self_.count_cycle,
            stress_range=stress_ranges_corr,
            mean_stress=mean_stress_corr,
            timestamp=self_.timestamp,
            range_bin_lower_bound=self_.range_bin_lower_bound,
            range_bin_width=self_.range_bin_width,
            _mean_bin_lower_bound=self_.mean_bin_lower_bound,
            mean_bin_width=self_.mean_bin_width,
            mean_stress_corrected="".join(
                f"{correction_type.upper()}: {ns.__dict__}"
            ),
            lffd_solved=self_.lffd_solved,
            name=self_.name,
            nr_small_cycles=self_.nr_small_cycles,
            residuals_sequence=self_.residuals_sequence,
            stress_concentration_factor=self_.stress_concentration_factor,
            _time_sequence=self_.time_sequence,
        )

        if len(self.time_sequence) == 1:
            w_msg = "".join(
                (
                    "If mean stress correction is performed in the contest ",
                    "of long-term fatigue analysis, please perform the sum ",
                    "of multiple CycleCount instances prior mean stress ",
                    "correction (MSC).\nIn fact, applying the MSC before ",
                    "summing into long-term CycleCount instance results in ",
                    "non-conservative life estimates, as after MSC, ",
                    "low-frequency fatigue cannot be estimated accurately.",
                )
            )

            warnings.warn(w_msg, UserWarning)

        return cc_msc

    # ! Plotting methods
    def plot_histogram(
        self,
        plot_type: str = "min-max",
        mode: str = "mountain",
        bins: Union[int, Sequence[int]] = 50,
        bin_range: Optional[np.ndarray] = None,
        normed: bool = False,
        weights: Optional[np.ndarray] = None,  # np.histogram2d args
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        dens_func: Optional[Callable] = None,
        **kwargs: dict,
    ) -> tuple:  # pragma: no cover
        """Plot the rainflow histogram of the :term:`Cycle-counting` instance
        with min and max stress on the x-y axes.

        Parameters
        ----------
        plot_type : str
            Type of plot to be generated, default is "min-max". Possible \
            values are: ["min-max", "mean-range"]
        mode : [None | 'mountain' | 'valley' | 'clip']
        Possible values are:
            - None : The points are plotted as one scatter object, in the
                order in-which they are specified at input
            - 'mountain' : The points are sorted/plotted in the order of
                the number of points in their 'bin'. This means that points
                in the highest density will be plotted on-top of others. This
                cleans-up the edges a bit, the points near the edges will
                overlap
            - 'valley' : The reverse order of 'mountain'. The low density
                bins are plotted on top of the high ones

        bins : int or array_like or [int, int] or [array, array]
            The bin specification, by default 50 bins are used:
                - If int, the number of bins for the two dimensions
                    (nx=ny=bins)
                - If array_like, the bin edges for the two dimensions
                    (x_edges = y_edges = bins)
                - If [int, int], the number of bins in each dimension
                    (n_x, n_y = bins)
                - If [array, array], the bin edges in each dimension
                    (x_edges, y_edges = bins)
                - A combination [int, array] or [array, int], where int
                    is the number of bins and array is the bin edges

        bin_range : array_like, shape(2,2), optional
            The leftmost and rightmost edges of the bins along each dimension
            (if not specified explicitly in the `bins` parameters):
            ``[[x_min, x_max], [y_min, y_max]]``. \
            All values outside of this range will be considered outliers
            and not tallied in the histogram.
            The default value is None, in which case all values are tallied
        normed : bool, optional
            If False, returns the number of samples in each bin. If True,
            returns the bin density ``bin_count / sample_count / bin_area``,
            default is False.
        weights : array_like, shape(N,), optional
            An array of values ``w_i`` weighing each sample ``(y_i, x_i)``.
            Weights are normalized to 1 if `normed` is True. If `normed` is
            False, the values of the returned histogram are equal to the sum of
            the weights belonging to the samples falling into each bin.
            The default is None, which gives each sample equal weight
        fig : matplotlib.figure.Figure, optional
            a Figure instance to add axes into
        dens_func : function or callable, optional
            A function that modifies (inputs and returns) the dens
            values (e.g., np.log10). The default is to not modify the
            values
        kwargs : these keyword arguments are all passed on to scatter

        Returns
        -------
        figure, paths : (
            `~matplotlib.figure.Figure`,
            `~matplotlib.axes.Axes`
        )
            The figure with scatter instance.
        """
        x = np.hstack([self.full_cycles[:, 0], self.residuals[:, 0]])
        y = np.hstack([self.full_cycles[:, 1], self.residuals[:, 1]])
        plot_types = [
            "min-max",
            "mean-range",
            "counts-range",
            "counts-range-cumsum",
            "counts-range-2D",
        ]
        if plot_type not in plot_types:
            e_msg = f"Invalid plot type. Must be one of {plot_types}."
            raise ValueError(e_msg)
        im = None
        if plot_type == "min-max":
            __tmp_x = x - y / 2
            y = x + y / 2
            x = __tmp_x
            bins = bins if bins is not None else 50
        if "counts-range" in plot_type:
            sor = sorted(
                zip(self.stress_range, self.count_cycle),
                reverse=True,
            )

            xy = np.asarray(list(map(list, sor)))
            x = xy[:, 1]
            y = xy[:, 0]
            bins = bins if bins is not None else self.bin_edges
            fig, axes = ht.make_axes(fig=fig, ax=ax)
            im = axes.scatter(
                np.cumsum(x) if "cumsum" in plot_type else x,
                y,
                # edgecolors="#CCC",
                # linewidths=0.5,
                c=x,
                **kwargs,  # type: ignore
            )
            # axes.set_yscale("log")
            # axes.set_xscale("log")
        if plot_type in ("min-max", "mean-range"):
            fig, axes, im = ht.xy_hist2d(
                x=x,
                y=y,
                mode=mode,
                bins=bins,
                bin_range=bin_range,
                normed=normed,
                weights=weights,
                dens_func=dens_func,
                fig=fig,
                ax=ax,
                **kwargs,
            )
            axes.axis("equal")
        cbar_label = (
            "# of cycles"
            if dens_func is None
            else "# of cycles(" + str(dens_func) + ")"
        )
        axes.set_xlabel(
            f"Min stress, {self.unit}"
            if plot_type == "min-max"
            else (
                f"Mean stress, {self.unit}"
                if plot_type == "mean-range"
                else "# of cycles"
            )
        )
        axes.set_ylabel(
            f"Max stress, {self.unit}"
            if plot_type == "min-max"
            else f"Range, {self.unit}"
        )
        cbar = plt.colorbar(im, ax=axes)
        cbar.set_label(cbar_label, rotation=270)
        cbar.ax.get_yaxis().labelpad = 10

        return fig, axes

    def plot_residuals_sequence(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        **kwargs,
    ) -> Tuple[
        matplotlib.figure.Figure, matplotlib.axes.Axes
    ]:  # pragma: no cover
        """Plot the residuals of the stress-strain curve.

        Parameters
        ----------
        fig : Optional[matplotlib.figure.Figure], optional
            a Figure instance to add axes into, by default None
        ax : Optional[matplotlib.axes.Axes], optional
            an Axes instance to add axes into, by default None

        Returns
        -------
        Tuple[
            matplotlib.figure.Figure,
            matplotlib.axes.Axes
        ]
            The figure with plot instance.
        """
        _, res_res_seq, res_res_idx = calc_rainflow(
            data=np.asarray(self.residuals_sequence),
            extended_output=True,
        )

        fig, axes = ht.make_axes(fig=fig, ax=ax)
        axes.plot(self.residuals_sequence, "#003366", ls=":", **kwargs)
        axes.plot(res_res_idx, res_res_seq, "#FF0080", **kwargs)
        axes.set_xlabel("Residuals sequence")
        axes.set_ylabel("Residuals")
        axes.grid(visible=True, which="major", color="#CCCCCC", linestyle="-")
        axes.minorticks_on()
        axes.grid(visible=True, which="minor", color="#E7E6DD", linestyle=":")
        axes.xaxis.grid(False, which="minor")
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)
        return fig, axes

    def plot_half_cycles_sequence(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        **kwargs,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """A symlink to plot_residuals_sequence.

        Parameters
        ----------
        fig : Optional[matplotlib.figure.Figure], optional
            a Figure instance to add axes into, by default None
        ax : Optional[matplotlib.axes.Axes], optional
            an Axes instance to add axes into, by default None

        Returns
        -------
        Tuple[
            matplotlib.figure.Figure,
            matplotlib.axes.Axes
        ]
            The figure with plot instance.
        """
        return self.plot_residuals_sequence(fig=fig, ax=ax, **kwargs)


def _multiplication_by_scalar(
    self_: "CycleCount", other_: Union[float, int]
) -> "CycleCount":
    """Multiplication of :class:`CycleCount` by a scalar factor

    Parameters
    ----------
    self_ : CycleCount
        CycleCount instance
    other_ : Union[float, int]
        Factor to multiply the instance by

    Returns
    -------
    CycleCount
        CycleCount instance with "raised" stresses and stress concentration
        factor != 1

    Raises
    ------
    TypeError
        Multiplication by a scalar is the only thing allowed
    """
    if not isinstance(other_, (float, int)):
        e_msg = "Multiplication by a scalar SCF is the only thing allowed"
        raise TypeError(e_msg)

    if isinstance(other_, int):
        other_ = float(other_)

    if self_.stress_concentration_factor != 1.0:
        w_msg = "".join(
            (
                f"SCF already defined and set to {TermColors.CGREEN2}",
                f"{self_.stress_concentration_factor}{TermColors.CYELLOW2}.",
            )
        )
        warnings.warn(w_msg, UserWarning)

    mul_range = other_ * self_.stress_range
    mul_mean = other_ * self_.mean_stress
    mul_res_seq = (
        other_ * np.array(self_.residuals_sequence)
        if self_.residuals_sequence is not None
        else np.empty(0)
    )
    mul_min_max = (
        other_ * np.array(self_.min_max_sequence)
        if self_.min_max_sequence is not None
        else np.empty(0)
    )
    mul_mean_bin = (
        other_ * self_.mean_bin_lower_bound
        if self_.mean_bin_lower_bound is not None
        else None
    )
    mul_scf = other_ * self_.stress_concentration_factor

    mul_cc = CycleCount(
        count_cycle=self_.count_cycle,
        mean_stress=mul_mean,
        stress_range=mul_range,
        range_bin_lower_bound=other_ * self_.range_bin_lower_bound,
        range_bin_width=other_ * self_.range_bin_width,
        _mean_bin_lower_bound=mul_mean_bin,
        mean_bin_width=other_ * self_.mean_bin_width,
        residuals_sequence=mul_res_seq,
        _min_max_sequence=mul_min_max,
        lffd_solved=self_.lffd_solved,
        mean_stress_corrected=self_.mean_stress_corrected,
        stress_concentration_factor=mul_scf,
        timestamp=self_.timestamp,
        name=str(self_.name) + ", SCF=" + str(other_),
    )

    # mul_cc.stress_concentration_factor = (
    #     other_ * self_.stress_concentration_factor
    # )
    return mul_cc


def _cycle_count_add_checks(
    self_: CycleCount,
    other_: CycleCount,
) -> bool:
    """Check if the two cycle count instances are compatible for addition.

    Parameters
    ----------
    self_ : CycleCount
        First cycle count instance
    other_ : Optional[CycleCount]
        Second cycle count instance

    Returns
    -------
    bool
        True if the two cycle count instances are compatible for addition
    """
    checks_ok = True
    # ! 1) TypeError
    # Trying to add to a non CycleCount object instance
    if self_.__class__.__name__ != other_.__class__.__name__:
        if (
            other_ is None
            or other_ == 0
            or isinstance(other_, float)
            and np.isnan(other_)
        ):
            w_msg = "".join(
                (
                    "Trying to add to a non CycleCount object instance ",
                    f"after {self_.timestamp.strftime('%d %B %Y, %H:%M')}.",
                )
            )
            warnings.warn(w_msg, UserWarning)
            return False
        raise TypeError(
            "Trying to add to a non CycleCount object instance",
            type(self_),
            type(other_),
        )
    # ! 2) TypeError
    # Trying to add-up different parameters.
    if self_.name != other_.name:
        w_msg = "".join(
            ("Summing different parameters. " f"({self_.name} + {other_.name})")
        )
        warnings.warn(w_msg, UserWarning)

    # ! 3) TypeError
    # Trying to add-up different units.
    if self_.unit != other_.unit:
        raise TypeError(
            "Trying to add-up CycleCounts having different units",
            self_.unit,
            other_.unit,
        )

    # ! 4) UserWarning
    # Different bin widths
    _bin_widths_add_check(self_.range_bin_width, other_.range_bin_width)
    _bin_widths_add_check(self_.mean_bin_width, other_.mean_bin_width, "mean")

    # ! 5) TypeError
    # Timestamps must be sorted ascending.
    if self_.timestamp is not None and other_.timestamp is not None:
        if np.max(self_.time_sequence) > np.min(other_.time_sequence):
            e_msg = (
                np.min(other_.time_sequence).strftime("%d %b, %Y, %H:%M")
                + " predates "
                + np.max(self_.time_sequence).strftime("%d %b, %Y, %H:%M")
                + ". Timestamps must be sorted ascending."
            )
            raise TypeError(e_msg)

    # ! 6) UserWarning
    # Adding-up instances having different stress concentration factors.
    if self_.stress_concentration_factor != other_.stress_concentration_factor:
        w_msg = "".join(
            (
                "Adding-up instances having different SCFs (",
                f"{TermColors.CGREEN}{self_.stress_concentration_factor} and ",
                f"{other_.stress_concentration_factor}{TermColors.CYELLOW2}",
                ").",
            )
        )
        warnings.warn(w_msg, UserWarning)

    return checks_ok


def _bin_widths_add_check(
    self_: float, other_: float, bin_range: str = "range"
) -> None:
    """Check if the two bin ranges are compatible for addition.

    Parameters
    ----------
    self_ : float
        First bin range
    other_ : float
        Second bin range
    bin_range : str, optional
        Bin range type. Default is "range".

    Returns
    -------
    None
    """

    if self_ != other_:
        w_msg = "".join(
            (
                f"Different bin {bin_range} widths. ({TermColors.CGREEN}",
                f"{self_} and {other_}{TermColors.CYELLOW2}).",
            )
        )
        warnings.warn(w_msg, UserWarning)


def _handling_different_bins_in_sum(
    self_: CycleCount,
    other_: "CycleCount",
) -> tuple:
    """Handle the different bin ranges for addition.

    Parameters
    ----------
    self_ : CycleCount
        First cycle count instance
    other_ : CycleCount
        Second cycle count instance

    Returns
    -------
    tuple
        Tuple containing the bin low bound and widths for mean and range
    """
    if self_.mean_bin_width is not None:
        if other_.mean_bin_width is not None:
            added_mean_bin_lower_bound = np.min(
                [self_.mean_bin_lower_bound, other_.mean_bin_lower_bound]
            )
            added_mean_bin_width = np.max(
                [self_.mean_bin_width, other_.mean_bin_width]
            )
        else:
            added_mean_bin_lower_bound = self_.mean_bin_lower_bound
            added_mean_bin_width = self_.mean_bin_width
    else:
        added_mean_bin_lower_bound = None
        added_mean_bin_width = None

    added_range_bin_lower_bound = np.min(
        [self_.range_bin_lower_bound, other_.range_bin_lower_bound]
    )
    added_range_bin_width = np.max(
        [self_.range_bin_width, other_.range_bin_width]
    )

    return (
        added_mean_bin_lower_bound,
        added_mean_bin_width,
        added_range_bin_lower_bound,
        added_range_bin_width,
    )


def _solve_lffd(self_: CycleCount, solve_mode: str = "Residuals") -> CycleCount:
    """Retrieve the LFFD for the cycle count instance.

    Parameters
    ----------
    self_ : CycleCount
        Cycle count instance
    solve_mode : str, optional
        Solve mode. Default is "Residuals".

    Returns
    -------
    CycleCount
        Cycle count instance with the LFFD solved.
    """

    ret_self, res_seq = _lffd_checks(self_, solve_mode)
    if ret_self:
        return self_

    # NOTE: next plot has been included for testing purposes

    res_rf, res_res_seq, _ = calc_rainflow(
        data=res_seq,
        extended_output=True,
    )

    cc_res = CycleCount(
        count_cycle=res_rf[:, 2],
        stress_range=2 * res_rf[:, 0],
        mean_stress=res_rf[:, 1],
        timestamp=self_.timestamp,
        _mean_bin_lower_bound=self_.mean_bin_lower_bound,
        mean_bin_width=self_.mean_bin_width,
        range_bin_lower_bound=self_.range_bin_lower_bound,
        range_bin_width=self_.range_bin_width,
        nr_small_cycles=self_.nr_small_cycles,
        residuals_sequence=res_res_seq,
        _min_max_sequence=(
            self_.min_max_sequence
            if "residuals" in solve_mode.lower()
            else np.asarray([min(res_res_seq), max(res_res_seq)])
        ),
        lffd_solved=True,
        mean_stress_corrected=self_.mean_stress_corrected,
        name=self_.name,
        _time_sequence=self_.time_sequence,
        stress_concentration_factor=self_.stress_concentration_factor,
    )

    if max(self_.count_cycle) <= 0.5:
        return cc_res

    cc_no_res = CycleCount(
        count_cycle=self_.count_cycle[self_.count_cycle > 0.5],
        stress_range=self_.stress_range[self_.count_cycle > 0.5],
        mean_stress=self_.mean_stress[self_.count_cycle > 0.5],
        timestamp=self_.timestamp,
        _mean_bin_lower_bound=self_.mean_bin_lower_bound,
        mean_bin_width=self_.mean_bin_width,
        range_bin_lower_bound=self_.range_bin_lower_bound,
        range_bin_width=self_.range_bin_width,
        residuals_sequence=np.empty(0),
        _min_max_sequence=np.empty(0),
        lffd_solved=True,
        mean_stress_corrected=self_.mean_stress_corrected,
        name=self_.name,
        _time_sequence=np.array(np.max(self_.time_sequence)),
        stress_concentration_factor=self_.stress_concentration_factor,
    )
    cc_solved = cc_res + cc_no_res
    cc_solved.lffd_solved = True
    return cc_solved


def _lffd_checks(self_: CycleCount, solve_mode: str) -> tuple:
    """Run checks on the :term:`LFFD` (Low-Frequency Fatigue Dynamics)
    retrieve method and sets the residuals sequence either to residuals
    or to min-max (approximated).

    Parameters
    ----------
    self_ : CycleCount
        The CycleCount object to check.
    solve_mode : str
        The solve mode to use.

    Returns
    -------
    tuple
        A tuple containing the following:
        - bool, True if the checks are passed, False otherwise.
        - np.ndarray, the residuals or min-max sequence.

    Raises
    ------
    ValueError
        If the solve mode is not supported.
    """

    solve_modes = [
        "min-max",
        "min/max",
        "min max",
        "residuals",
    ]
    if solve_mode.lower() not in solve_modes:
        e_msg = f"Invalid solve mode. Must be one of {solve_modes}."
        raise ValueError(e_msg)

    return_self = False
    res_sequence = np.empty(0)
    if self_.lffd_solved:
        w_msg = "Residuals already resolved. Nothing to do."
        warnings.warn(w_msg, UserWarning)
        return_self = True
    if (
        hasattr(self_, "residuals_sequence")
        and self_.residuals_sequence is not None
    ):
        if "residuals" in solve_mode.lower():
            res_sequence = self_.residuals_sequence
        if "min" in solve_mode.lower() or "max" in solve_mode.lower():
            res_sequence = self_.min_max_sequence
    else:
        w_msg = "'residuals_sequence' is not available. Nothing to do."
        warnings.warn(w_msg, UserWarning)
        return_self = True
    if len(res_sequence) < 3:
        w_msg = "".join(
            (
                f"lffd_solve() not possible when{TermColors.CGREEN2}",
                f"{TermColors.CBOLD}len{TermColors.CEND}{TermColors.CYELLOW2}("
                f"residuals_sequence) < 3.\n Returning {TermColors.CBEIGE2}",
                f"{TermColors.CBOLD}'self'{TermColors.CEND}.",
            )
        )
        warnings.warn(w_msg, UserWarning)
        return_self = True
    return return_self, res_sequence


def pbar_sum(cc_list: Sequence[CycleCount]) -> CycleCount:
    """Sum a list of CycleCount objects using a progress bar.

    Parameters
    ----------
    cc_list : list
        The list of CycleCount objects to sum.

    Returns
    -------
    py_fatigue.CycleCount
        The sum of the CycleCount objects.
    """

    if len(cc_list) == 1:
        return cc_list[0]

    chunk_size = int(np.sqrt(len(cc_list)))
    cc_list = [cc for cc in cc_list if cc is not None]

    sleep(0.15)
    chunked_cc_list = list(pfu.chunks(cc_list, chunk_size))
    l_c = len(chunked_cc_list)

    cc_sum: CycleCount = chunked_cc_list[0]  # type: ignore
    cc_sum_list: list[CycleCount] = []

    for j, chunked_cc_sublist in enumerate(chunked_cc_list):
        k = j + 1
        # try:
        blocks = int(k / l_c * 32)
        # except ZeroDivisionError:
        #     blocks = int(k / l_c * 32)
        underscores = 32 - blocks
        pbar = f"{k * len(chunked_cc_sublist)}/{len(cc_list)} -> "
        pbar += TermColors.CBLUE2 + "|" + "".join(["█"] * blocks)
        pbar += "".join(["_"] * underscores) + "|" + TermColors.CEND
        # try:
        pbar += f" -> {min(int(k / l_c * 100), 100)}% completed... "
        # except ZeroDivisionError:
        #     pbar += f" -> { min(int(k / l_c * 100), 100)}% completed... "
        print(pbar, end="\r")

        if len(chunked_cc_sublist) == 1:
            cc_sum_list.append(chunked_cc_sublist[0])
            cc_sum = (
                cc_sum_list[-1]
                if len(cc_sum_list) == 1
                else cc_sum + cc_sum_list[-1]
            )
            continue
        cc_sum_list.append(chunked_cc_sublist[0])
        for cc in chunked_cc_sublist[1::]:
            if cc is None:
                continue
            cc_sum_list[-1] += cc
        cc_sum = (
            cc_sum_list[-1]
            if len(cc_sum_list) == 1
            else cc_sum + cc_sum_list[-1]
        )
    return cc_sum



"""The module :mod:`py_fatigue.stress_range.stress_range` contains
the main stress_range-related utilities.
"""

from typing import ClassVar

import numpy as np

from ..utils import FatigueStress

__all__ = ["StressRange"]


class StressRange(FatigueStress):
    """Stress range class."""

    category: ClassVar[str] = "range"

    def small_cycles_idx(self):
        """The index of the small cycles."""
        return np.where(self.full >= self.bin_lb)[0]



# -*- coding: utf-8 -*-
"""Py-fatigue bundles the main functionality for performing cyclic stress (fatigue) analysis and cycle-counting."""  # noqa: E501  # pylint: disable=C0301
from .cycle_count.cycle_count import CycleCount
from .material.sn_curve import SNCurve
from .material.crack_growth_curve import ParisCurve, WalkerCurve
from .version import __version__
from . import cycle_count, geometry, material, damage, styling, testing

__all__ = [
    "CycleCount",
    "SNCurve",
    "ParisCurve",
    "WalkerCurve",
    "cycle_count",
    "damage",
    "geometry",
    "material",
    "styling",
    "testing",
    "__version__",
]



"""The module :mod:`py_fatigue.mean_stress.mean_stress` contains
the main mean stress-related utilities.
"""

from __future__ import annotations
from typing import ClassVar

import numpy as np

from ..utils import FatigueStress


class MeanStress(FatigueStress):
    """Mean stress class."""

    category: ClassVar[str] = "mean"


def calculate_mean_stress(
    load_ratio: float | np.ndarray | list,
    stress_amplitude: float | np.ndarray | list,
) -> float | np.ndarray:
    """
    Calculate the mean stress based on the load ratio and stress amplitude.

    The relationship is:

        mean_stress = stress_amplitude * (1 + load_ratio) / 2

    Parameters
    ----------
    load_ratio : float | np.ndarray | list
        Load ratio (minimum stress / maximum stress).
    stress_amplitude : float | np.ndarray | list
        Stress amplitude (half the range of stress variation).

    Returns
    -------
    float | np.ndarray
        Calculated mean stress.
    """
    if isinstance(load_ratio, list):
        load_ratio = np.array(load_ratio)
    if isinstance(stress_amplitude, list):
        stress_amplitude = np.array(stress_amplitude)
    return stress_amplitude * (1 + load_ratio) / (1 - load_ratio)
